#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# flake8: noqa

# Must happen before any importing of tensorflow to curtail mem usage
from merlin.models.loader.tf_utils import configure_tensorflow
from merlin.models.tf.core.index import IndexBlock, TopKIndexBlock
from merlin.models.tf.core.tabular import AsTabular, Filter, TabularBlock

configure_tensorflow()

from tensorflow.keras.layers import Dense, Layer
from tensorflow.keras.losses import Loss
from tensorflow.keras.metrics import Metric
from tensorflow.keras.optimizers import Optimizer
from tensorflow.python.training.tracking.data_structures import ListWrapper, _DictWrapper

from merlin.models.loader.tf_utils import configure_tensorflow
from merlin.models.tf.blocks.cross import CrossBlock
from merlin.models.tf.blocks.dlrm import DLRMBlock
from merlin.models.tf.blocks.experts import CGCBlock, MMOEBlock, MMOEGate
from merlin.models.tf.blocks.interaction import (
    DotProductInteraction,
    FMBlock,
    FMPairwiseInteraction,
)
from merlin.models.tf.blocks.mlp import DenseResidualBlock, MLPBlock
from merlin.models.tf.blocks.optimizer import (
    LazyAdam,
    MultiOptimizer,
    OptimizerBlocks,
    split_embeddings_on_size,
)
from merlin.models.tf.blocks.retrieval.base import DualEncoderBlock, ItemRetrievalScorer
from merlin.models.tf.blocks.retrieval.matrix_factorization import (
    MatrixFactorizationBlock,
    QueryItemIdsEmbeddingsBlock,
)
from merlin.models.tf.blocks.retrieval.two_tower import TwoTowerBlock
from merlin.models.tf.blocks.sampling.base import ItemSampler
from merlin.models.tf.blocks.sampling.cross_batch import PopularityBasedSampler
from merlin.models.tf.blocks.sampling.in_batch import InBatchSampler
from merlin.models.tf.blocks.sampling.queue import FIFOQueue
from merlin.models.tf.core.aggregation import (
    ConcatFeatures,
    ElementwiseSum,
    ElementwiseSumItemMulti,
    StackFeatures,
)
from merlin.models.tf.core.base import (
    Block,
    EmbeddingWithMetadata,
    ModelContext,
    NoOp,
    right_shift_layer,
)
from merlin.models.tf.core.combinators import (
    Cond,
    MapValues,
    ParallelBlock,
    ResidualBlock,
    SequentialBlock,
)
from merlin.models.tf.core.encoder import EmbeddingEncoder, Encoder, TopKEncoder
from merlin.models.tf.inputs.base import InputBlock, InputBlockV2
from merlin.models.tf.inputs.continuous import Continuous, ContinuousFeatures, ContinuousProjection
from merlin.models.tf.inputs.embedding import (
    AverageEmbeddingsByWeightFeature,
    ContinuousEmbedding,
    EmbeddingFeatures,
    EmbeddingOptions,
    Embeddings,
    EmbeddingTable,
    FeatureConfig,
    SequenceEmbeddingFeatures,
    TableConfig,
)
from merlin.models.tf.loader import KerasSequenceValidator, Loader, sample_batch
from merlin.models.tf.losses import LossType
from merlin.models.tf.metrics.topk import (
    AvgPrecisionAt,
    MRRAt,
    NDCGAt,
    PrecisionAt,
    RecallAt,
    TopKMetricsAggregator,
)
from merlin.models.tf.models import benchmark
from merlin.models.tf.models.base import BaseModel, Model, RetrievalModel, RetrievalModelV2
from merlin.models.tf.models.ranking import DCNModel, DeepFMModel, DLRMModel, WideAndDeepModel
from merlin.models.tf.models.retrieval import (
    MatrixFactorizationModel,
    TwoTowerModel,
    YoutubeDNNRetrievalModel,
)
from merlin.models.tf.outputs.base import ModelOutput
from merlin.models.tf.outputs.classification import BinaryOutput, CategoricalOutput
from merlin.models.tf.outputs.contrastive import ContrastiveOutput
from merlin.models.tf.outputs.regression import RegressionOutput
from merlin.models.tf.outputs.sampling.base import Candidate, CandidateSampler
from merlin.models.tf.outputs.sampling.in_batch import InBatchSamplerV2
from merlin.models.tf.outputs.sampling.popularity import PopularityBasedSamplerV2
from merlin.models.tf.outputs.topk import TopKOutput
from merlin.models.tf.prediction_tasks.base import ParallelPredictionBlock, PredictionTask
from merlin.models.tf.prediction_tasks.classification import (
    BinaryClassificationTask,
    MultiClassClassificationTask,
)
from merlin.models.tf.prediction_tasks.multi import PredictionTasks
from merlin.models.tf.prediction_tasks.regression import RegressionTask
from merlin.models.tf.prediction_tasks.retrieval import ItemRetrievalTask
from merlin.models.utils.dependencies import is_transformers_available

if is_transformers_available():
    from merlin.models.tf.transformers.block import (
        AlbertBlock,
        BertBlock,
        GPT2Block,
        RobertaBlock,
        TransformerBlock,
        XLNetBlock,
    )
    from merlin.models.tf.transformers.transforms import (
        AttentionWeights,
        HiddenStates,
        LastHiddenState,
        LastHiddenStateAndAttention,
    )

from merlin.models.tf.transforms.features import (
    CategoryEncoding,
    HashedCross,
    HashedCrossAll,
    ToDense,
    ToOneHot,
    ToSparse,
    ToTarget,
)
from merlin.models.tf.transforms.noise import StochasticSwapNoise
from merlin.models.tf.transforms.regularization import L2Norm
from merlin.models.tf.transforms.sequence import (
    ReplaceMaskedEmbeddings,
    SequenceMaskLast,
    SequenceMaskRandom,
    SequencePredictLast,
    SequencePredictNext,
    SequencePredictRandom,
    SequenceTargetAsInput,
)
from merlin.models.tf.transforms.tensor import ExpandDims, ListToDense, ListToRagged, ListToSparse
from merlin.models.tf.utils import repr_utils
from merlin.models.tf.utils.tf_utils import TensorInitializer

ListWrapper.__repr__ = repr_utils.list_wrapper_repr
_DictWrapper.__repr__ = repr_utils.dict_wrapper_repr

Dense.repr_extra = repr_utils.dense_extra_repr
Layer.__rrshift__ = right_shift_layer
Layer.__repr__ = repr_utils.layer_repr
Loss.__repr__ = repr_utils.layer_repr_no_children
Metric.__repr__ = repr_utils.layer_repr_no_children
Optimizer.__repr__ = repr_utils.layer_repr_no_children

__all__ = [
    "Block",
    "Cond",
    "MapValues",
    "ModelContext",
    "SequentialBlock",
    "ResidualBlock",
    "DualEncoderBlock",
    "TopKEncoder",
    "Encoder",
    "EmbeddingEncoder",
    "CrossBlock",
    "DLRMBlock",
    "MLPBlock",
    "ContinuousEmbedding",
    "MMOEGate",
    "MMOEBlock",
    "CGCBlock",
    "TopKIndexBlock",
    "IndexBlock",
    "DenseResidualBlock",
    "TabularBlock",
    "ContinuousFeatures",
    "Continuous",
    "ContinuousProjection",
    "EmbeddingFeatures",
    "SequenceEmbeddingFeatures",
    "EmbeddingOptions",
    "EmbeddingTable",
    "AverageEmbeddingsByWeightFeature",
    "Embeddings",
    "FeatureConfig",
    "TableConfig",
    "ParallelPredictionBlock",
    "TwoTowerBlock",
    "MatrixFactorizationBlock",
    "QueryItemIdsEmbeddingsBlock",
    "ListToDense",
    "ListToRagged",
    "ListToSparse",
    "ToSparse",
    "ToDense",
    "ToTarget",
    "CategoryEncoding",
    "HashedCross",
    "HashedCrossAll",
    "ElementwiseSum",
    "ElementwiseSumItemMulti",
    "AsTabular",
    "ConcatFeatures",
    "Filter",
    "ParallelBlock",
    "StackFeatures",
    "DotProductInteraction",
    "FMPairwiseInteraction",
    "FMBlock",
    "ToOneHot",
    "ModelOutput",
    "BinaryOutput",
    "RegressionOutput",
    "CategoricalOutput",
    "ContrastiveOutput",
    "PredictionTask",
    "BinaryClassificationTask",
    "MultiClassClassificationTask",
    "RegressionTask",
    "MultiOptimizer",
    "LazyAdam",
    "OptimizerBlocks",
    "split_embeddings_on_size",
    "OptimizerBlocks",
    "ItemRetrievalTask",
    "ItemRetrievalScorer",
    "NDCGAt",
    "PrecisionAt",
    "MRRAt",
    "AvgPrecisionAt",
    "RecallAt",
    "BaseModel",
    "TopKMetricsAggregator",
    "Model",
    "RetrievalModel",
    "RetrievalModelV2",
    "InputBlock",
    "InputBlockV2",
    "PredictionTasks",
    "StochasticSwapNoise",
    "ExpandDims",
    "L2Norm",
    "NoOp",
    "ItemSampler",
    "EmbeddingWithMetadata",
    "InBatchSampler",
    "PopularityBasedSampler",
    "FIFOQueue",
    "YoutubeDNNRetrievalModel",
    "TwoTowerModel",
    "MatrixFactorizationModel",
    "DLRMModel",
    "DCNModel",
    "DeepFMModel",
    "WideAndDeepModel",
    "losses",
    "LossType",
    "sample_batch",
    "TensorInitializer",
    "BroadcastToSequence",
    "Loader",
    "KerasSequenceValidator",
    "SequencePredictNext",
    "SequencePredictLast",
    "SequencePredictRandom",
    "SequenceTargetAsInput",
    "SequenceMaskLast",
    "SequenceMaskRandom",
    "ReplaceMaskedEmbeddings",
]



import numpy as np
import pytest
import tensorflow as tf
from tensorflow.test import TestCase

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.utils import testing_utils
from merlin.schema import Tags


@pytest.mark.parametrize("name_branches", [True, False])
def test_parallel_block_pruning(music_streaming_data: Dataset, name_branches: bool):
    music_streaming_data.schema = music_streaming_data.schema.remove_by_tag(Tags.CONTINUOUS)

    continuous_block = mm.Filter(music_streaming_data.schema.select_by_tag(Tags.CONTINUOUS))
    embedding_block = mm.Embeddings(music_streaming_data.schema.select_by_tag(Tags.CATEGORICAL))

    if name_branches:
        branches = {"continuous": continuous_block, "embeddings": embedding_block}
    else:
        branches = [continuous_block, embedding_block]

    input_block = mm.ParallelBlock(branches, schema=music_streaming_data.schema)

    features = mm.sample_batch(music_streaming_data, batch_size=10, include_targets=False)

    outputs = input_block(features)

    assert len(outputs) == 7  # There are 7 categorical features
    assert continuous_block not in input_block.parallel_values


def test_parallel_block_serialization(music_streaming_data: Dataset):
    unknown_filter = mm.Filter(["none"])
    block = mm.ParallelBlock(mm.Filter(["position"]), unknown_filter, automatic_pruning=False)
    block_copy = block.from_config(block.get_config())

    assert not block_copy.automatic_pruning
    assert unknown_filter not in block_copy.parallel_values

    features = mm.sample_batch(music_streaming_data, batch_size=10, include_targets=False)

    outputs_1 = block(features)
    outputs_2 = block_copy(features)

    for key in outputs_1:
        np.testing.assert_array_equal(outputs_2[key].numpy(), outputs_1[key].numpy())
    assert len(outputs_1) == len(outputs_2) == 1


@pytest.mark.parametrize("name_branches", [True, False])
def test_parallel_block_schema_propagation(music_streaming_data, name_branches: bool):
    continuous_block = mm.Filter(Tags.CONTINUOUS)
    embedding_block = mm.Embeddings(music_streaming_data.schema.select_by_tag(Tags.CATEGORICAL))

    if name_branches:
        branches = {"continuous": continuous_block, "embeddings": embedding_block}
    else:
        branches = [continuous_block, embedding_block]

    input_block = mm.ParallelBlock(branches, schema=music_streaming_data.schema)
    features = mm.sample_batch(music_streaming_data, batch_size=10, include_targets=False)
    outputs = input_block(features)

    assert len(outputs) == 10  # There are 7 categorical + 3 continuous features


@pytest.mark.parametrize("name_branches", [True, False])
def test_parallel_block_with_layers(music_streaming_data, name_branches: bool):
    d, d_1 = tf.keras.layers.Dense(32), tf.keras.layers.Dense(32)
    if name_branches:
        branches = {"d": d, "d_1": d_1}
    else:
        branches = [d, d_1]

    block = mm.ParallelBlock(branches, aggregation="concat")
    model = mm.Model.from_block(block, music_streaming_data.schema)

    outputs = block(tf.constant([[2.0]]))
    assert outputs.shape == tf.TensorShape([1, 64])

    features = mm.sample_batch(music_streaming_data, batch_size=10, include_targets=False)
    outputs = model(features)

    assert len(outputs) == 3  # number of prediction tasks in this schema


def test_parallel_block_select_from_names():
    test_case = TestCase()
    branches = {
        "1": tf.keras.layers.Dense(32),
        "2": tf.keras.layers.Dense(32),
        "3": tf.keras.layers.Dense(32),
    }
    blocks = mm.ParallelBlock(branches, aggregation="concat")
    assert 2 == len(blocks.select_by_names(["1", "2"]))
    with test_case.assertRaisesRegex(ValueError, "is not in ParallelBlock"):
        blocks.select_by_names(["0", "2"])


def test_parallel_block_select_by_tags(music_streaming_data):
    continuous_block = mm.Filter(Tags.CONTINUOUS)
    embedding_block = mm.Embeddings(
        schema=music_streaming_data.schema.select_by_tag(Tags.CATEGORICAL)
    )
    branches = {"continuous": continuous_block, "embeddings": embedding_block}
    parallel_block = mm.ParallelBlock(branches, schema=music_streaming_data.schema)

    continuous_inputs = parallel_block.select_by_tag(Tags.CONTINUOUS)
    assert isinstance(continuous_inputs, mm.ParallelBlock)
    assert sorted(continuous_inputs.schema.column_names) == [
        "item_recency",
        "position",
        "user_age",
    ]

    categorical_inputs = parallel_block.select_by_tag(Tags.CATEGORICAL)
    assert sorted(categorical_inputs.schema.column_names) == [
        "country",
        "item_category",
        "item_genres",
        "item_id",
        "session_id",
        "user_genres",
        "user_id",
    ]

    # Forward pass a batch and compare the input features to output features
    batch = mm.sample_batch(music_streaming_data, batch_size=10, include_targets=False)
    continuous_outputs = continuous_inputs(batch)
    assert sorted(continuous_outputs.keys()) == sorted(continuous_inputs.schema.column_names)
    categorical_outputs = categorical_inputs(batch)
    assert sorted(categorical_outputs.keys()) == sorted(categorical_inputs.schema.column_names)

    # There are no sequence or time features in the dataset.
    assert parallel_block.select_by_tag([Tags.SEQUENCE, Tags.TIME]) is None

    # InputBlock is also a ParallelBlock.
    input_block = mm.InputBlockV2(music_streaming_data.schema)
    item_inputs = input_block.select_by_tag(Tags.ITEM)
    assert sorted(item_inputs.schema.column_names) == [
        "item_category",
        "item_genres",
        "item_id",
        "item_recency",
    ]
    outputs = item_inputs(batch)
    assert sorted(outputs.keys()) == sorted(item_inputs.schema.column_names)


class TestCond:
    def test_true(self):
        condition = tf.keras.layers.Lambda(lambda _: True)
        true = tf.keras.layers.Lambda(lambda _: tf.constant([4.0]))
        false = tf.keras.layers.Lambda(lambda _: tf.constant([2.0]))
        output_data = testing_utils.layer_test(
            mm.Cond, kwargs=dict(condition=condition, true=true, false=false), input_shape=(1,)
        )
        np.testing.assert_array_equal(output_data, np.array([4.0]))

    def test_false(self):
        condition = tf.keras.layers.Lambda(lambda _: False)
        true = tf.keras.layers.Lambda(lambda _: tf.constant([4.0]))
        false = tf.keras.layers.Lambda(lambda _: tf.constant([2.0]))
        output_data = testing_utils.layer_test(
            mm.Cond, kwargs=dict(condition=condition, true=true, false=false), input_shape=(1,)
        )
        np.testing.assert_array_equal(output_data, np.array([2.0]))

    def test_divide(self):
        condition = tf.keras.layers.Lambda(lambda _: True)
        true = tf.keras.layers.Lambda(lambda x: x / 2)
        false = tf.keras.layers.Lambda(lambda x: x / 5)
        output_data = testing_utils.layer_test(
            mm.Cond,
            kwargs=dict(condition=condition, true=true, false=false),
            input_data=tf.convert_to_tensor([np.arange(5).astype(np.float32)]),
        )
        np.testing.assert_array_equal(output_data, np.array([[0.0, 0.5, 1.0, 1.5, 2.0]]))

    def test_default_false(self):
        condition = tf.keras.layers.Lambda(lambda _: False)
        true = tf.keras.layers.Lambda(lambda _: tf.constant([4.0]))
        output_data = testing_utils.layer_test(
            mm.Cond, kwargs=dict(condition=condition, true=true), input_data=tf.constant([3.0])
        )
        np.testing.assert_array_equal(output_data, np.array([3.0]))

    def test_different_output_shapes(self):
        condition = tf.keras.layers.Lambda(lambda _: True)
        true = tf.keras.layers.Dense(1)
        false = tf.keras.layers.Dense(2)
        with pytest.raises(ValueError) as exc_info:
            testing_utils.layer_test(
                mm.Cond,
                kwargs=dict(condition=condition, true=true, false=false),
                input_data=tf.constant([[1.0]]),
            )
        assert "true and false branches must return the same output shape" in str(exc_info.value)

    def test_with_blocks(self):
        condition = tf.keras.layers.Lambda(lambda _: True)
        true = mm.MLPBlock([10])
        false = mm.SequentialBlock([mm.MLPBlock([100]), mm.MLPBlock([10])])
        output_data = testing_utils.layer_test(
            mm.Cond,
            kwargs=dict(condition=condition, true=true, false=false),
            input_data=tf.constant([[3.0]]),
        )
        assert output_data.shape == (1, 10)

    @pytest.mark.parametrize("run_eagerly", [True, False])
    def test_with_model(self, run_eagerly, music_streaming_data):
        condition = tf.keras.layers.Lambda(lambda _: tf.random.uniform((1,)) < 0.5)

        true = tf.keras.layers.Lambda(lambda x: x * 2)

        layer = mm.Cond(condition, true)

        model = mm.Model(
            tf.keras.layers.Lambda(lambda x: x["item_recency"]),
            layer,
            tf.keras.layers.Dense(1),
            mm.BinaryClassificationTask("click"),
        )

        testing_utils.model_test(model, music_streaming_data, run_eagerly=run_eagerly)


class TestMapValues:
    def test_map_values_tensor(self):
        map_layer = tf.keras.layers.Lambda(lambda _: tf.constant([4.0]))
        output_data = testing_utils.layer_test(
            mm.MapValues, kwargs=dict(layer=map_layer), input_shape=(1,)
        )
        np.testing.assert_array_equal(output_data, np.array([4.0]))

    def test_map_values_dict(self):
        map_layer = mm.MapValues(tf.keras.layers.Lambda(lambda _: tf.constant([4.0])))

        out = map_layer({"a": tf.constant([1.0]), "b": tf.constant([2.0])})
        np.testing.assert_array_equal(out, {k: np.array([4.0]) for k in ["a", "b"]})

    def test_map_values_in_model(self, music_streaming_data):
        map_layer = mm.MapValues(tf.keras.layers.Lambda(lambda x: x + 1))

        model = mm.Model(
            mm.InputBlock(music_streaming_data.schema),
            map_layer,
            mm.MLPBlock([10]),
            mm.BinaryClassificationTask("click"),
        )

        testing_utils.model_test(model, music_streaming_data)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from testbook import testbook

from tests.conftest import REPO_ROOT


@testbook(REPO_ROOT / "examples/01-Getting-started.ipynb", execute=False)
def test_example_01_getting_started(tb):
    tb.inject(
        """
        from unittest.mock import patch
        from merlin.datasets.synthetic import generate_data
        mock_train, mock_valid = generate_data(
            input="movielens-1m",
            num_rows=1000,
            set_sizes=(0.8, 0.2)
        )
        p1 = patch(
            "merlin.datasets.entertainment.get_movielens",
            return_value=[mock_train, mock_valid]
        )
        p1.start()
        """
    )
    tb.execute()
    metrics = tb.ref("metrics")
    assert set(metrics.keys()) == set(
        [
            "auc",
            "binary_accuracy",
            "loss",
            "regularization_loss",
            "precision",
            "recall",
        ]
    )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict

import tensorflow as tf
from tensorflow.keras.layers import Layer
from transformers.modeling_tf_outputs import TFBaseModelOutputWithPoolingAndCrossAttentions

from merlin.models.tf.core.base import Block


@Block.registry.register("last_hidden_state")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class LastHiddenState(Layer):
    """Select the hidden states of the last transform layer
    from the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return inputs.last_hidden_state


@Block.registry.register("pooler_output")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class PoolerOutput(Layer):
    """Select the pooled representation of the sequence
    from the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return inputs.pooler_output


@Block.registry.register("hidden_states")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class HiddenStates(Layer):
    """Select the tuple of hidden states of all transformer layers
    from the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return inputs.hidden_states


@Block.registry.register("attentions")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class AttentionWeights(Layer):
    """Select the attention weights of all transformer layers
    from the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return inputs.attentions


@Block.registry.register("last_hidden_state_and_attention")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class LastHiddenStateAndAttention(Layer):
    """Select the hidden states and attention weights of the last
    tranfrom the HF dataclass object
    `TFBaseModelOutputWithPoolingAndCrossAttentions`

    Parameters
    ----------
    inputs: TFBaseModelOutputWithPoolingAndCrossAttentions
        The output class returned by the HuggingFace transformer layer
    """

    def call(self, inputs: TFBaseModelOutputWithPoolingAndCrossAttentions):
        return (inputs.last_hidden_state, inputs.attentions[-1])


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class PrepareTransformerInputs(tf.keras.layers.Layer):
    """Prepare the dictionary of inputs expected by the transformer layer"""

    def call(self, inputs: tf.Tensor) -> Dict[str, tf.Tensor]:
        if isinstance(inputs, tf.RaggedTensor):
            # convert to a dense tensor as HF transformers do not support ragged tensors
            inputs = inputs.to_tensor()
        return {"inputs_embeds": inputs}



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import glob
import os
from pathlib import Path
from typing import Optional, Tuple, Union

import numpy as np

import merlin.io
from merlin.core.utils import download_file
from merlin.datasets import BASE_PATH
from merlin.models.utils.example_utils import workflow_fit_transform
from merlin.models.utils.nvt_utils import require_nvt
from merlin.schema import Tags

try:
    import nvtabular as nvt

    Workflow = nvt.Workflow
except ImportError:
    Workflow = None


def get_criteo(
    path: Union[str, Path] = None,
    num_days: int = 2,
    overwrite: bool = False,
    transformed_name: str = "transformed",
    nvt_workflow: Optional[Workflow] = None,
    valid_num_days: int = 1,
    **kwargs,
) -> Tuple[merlin.io.Dataset, merlin.io.Dataset]:
    """
    Get the Criteo 1TB Click Logs dataset, more info can be found
    [here](https://ailab.criteo.com/download-criteo-1tb-click-logs-dataset/).

    This dataset contains feature values and click feedback for millions of display
    ads. Its purpose is to benchmark algorithms for clickthrough rate (CTR) prediction.
    It is similar, but larger, to the dataset released for the Display Advertising
    Challenge hosted by Kaggle: https://www.kaggle.com/c/criteo-display-ad-challenge

    This function will download the data if not already present.
    In case you want to use this dataset to run our tutorials, you can also opt for synthetic data.
    Synthetic data can be generated by running::
        from merlin.models.data.synthetic import generate_data

        data = generate_data("criteo")

    Parameters
    ----------
    path: str or Path, optional
        Path to the dataset. If not set will default to
        the 'merlin-models-data` directory in your home folder
    num_days: int, optional
        Number of days to download, should be between 2 & 23. Default is 2.
        The last day will be used for the valid set, the rest for train.
    overwrite: bool, optional
        Whether to overwrite the transformed data.
    transformed_name: str, optional
        Name of the transformed data.
        This is useful when you want to load the transformed data
        multiple times using different workflows.
    nvt_workflow: Optional[Workflow]
        Workflow to transform the raw data.
        If None, the default workflow will be used from `default_criteo_transform`.
    valid_num_days: int.
        Number of days to use for the validation set. This will be the last day(s).
        Default is 1. This is similar to all papers published on the Criteo dataset.

    Returns
    -------
    train: merlin.io.Dataset
        Training dataset.
    valid: merlin.io.Dataset
        Test dataset.
    """

    require_nvt()

    if path is None:
        p = Path(BASE_PATH) / "criteo"
    else:
        p = Path(path)

    raw_data_path = p / "orig"
    maybe_download(raw_data_path, num_days)

    variant_path = p / f"{num_days}-days"
    raw_path = variant_path / "raw"
    if not variant_path.exists():
        raw_path.mkdir(parents=True)
        prepare_criteo(raw_data_path, raw_path, num_days=num_days)

    nvt_path = variant_path / transformed_name
    train_path, valid_path = nvt_path / "train", nvt_path / "valid"
    nvt_path_exists = train_path.exists() and valid_path.exists()
    if not nvt_path_exists or overwrite:
        transform_criteo(
            raw_path, variant_path, workflow=nvt_workflow, valid_num_days=valid_num_days, **kwargs
        )

    train = merlin.io.Dataset(str(train_path), engine="parquet")
    valid = merlin.io.Dataset(str(valid_path), engine="parquet")

    return train, valid


def maybe_download(destination: Union[str, Path], num_days: int):
    if num_days < 2 or num_days > 23:
        raise ValueError(
            str(num_days)
            + " is not supported. A minimum of 2 days are "
            + "required and a maximum of 24 (0-23 days) are available"
        )

    # Create input dir if not exists
    if not os.path.exists(destination):
        os.makedirs(destination)

    # Iterate over days
    for i in range(0, num_days):
        file = os.path.join(destination, "day_" + str(i) + ".gz")
        # Download file, if there is no .gz, .csv or .parquet file
        if not (
            os.path.exists(file)
            or os.path.exists(
                file.replace(".gz", ".parquet").replace("crit_orig", "converted/criteo/")
            )
            or os.path.exists(file.replace(".gz", ""))
        ):
            download_file(
                "https://storage.googleapis.com/criteo-cail-datasets/day_" + str(i) + ".gz",
                file,
            )


def prepare_criteo(
    raw_data_path: Union[str, Path],
    output_path: Union[str, Path],
    num_days: int,
    part_mem_fraction=0.1,
    **kwargs,
):
    """Convert the raw data to parquet format.

    Parameters
    ----------
    raw_data_path: str or Path
        Path to the raw data.
    output_path: str or Path
        Path to the output directory.
    num_days: int
        Number of days to use.
    part_mem_fraction: float, optional
        Fraction of memory to use for the conversion.
    client:

    Returns
    -------
    output_path: str or Path
    """

    if not os.path.exists(output_path):
        os.makedirs(output_path)

    # Specify column names
    cont_names = ["I" + str(x) for x in range(1, 14)]
    cat_names = ["C" + str(x) for x in range(1, 27)]
    cols = ["label"] + cont_names + cat_names

    # Specify column dtypes. Note that "hex" means that
    # the values will be hexadecimal strings that should
    # be converted to int32
    dtypes = {}
    dtypes["label"] = np.int32
    for x in cont_names:
        dtypes[x] = np.int32
    for x in cat_names:
        dtypes[x] = "hex"

    # Create an NVTabular Dataset from a CSV-file glob
    file_list = glob.glob(os.path.join(str(raw_data_path), "day_*[!.gz]"))
    file_list = sorted(file_list)[:num_days]
    dataset = nvt.Dataset(
        file_list,
        engine="csv",
        names=cols,
        part_mem_fraction=part_mem_fraction,
        sep="\t",
        dtypes=dtypes,
        **kwargs,
    )

    dataset.to_parquet(str(output_path), preserve_files=True)

    return output_path


def default_criteo_transform(num_buckets=10000000, **kwargs) -> Workflow:
    from nvtabular import ops as nvt_ops

    continuous = ["I" + str(x) for x in range(1, 14)]
    categorical = ["C" + str(x) for x in list(range(1, 21)) + list(range(22, 27))]
    # It's a reasonable guess that C21 is the item-id col.
    # This is only used to generate synthetic data
    item_id = ["C21"] >> nvt_ops.AddMetadata(tags=[Tags.ITEM_ID])
    targets = ["label"] >> nvt_ops.AddMetadata(tags=["target", Tags.BINARY_CLASSIFICATION])

    categorify_op = nvt_ops.Categorify(max_size=num_buckets)
    cat_features = categorical + item_id >> categorify_op
    cont_features = (
        continuous >> nvt_ops.FillMissing() >> nvt_ops.Clip(min_value=0) >> nvt_ops.Normalize()
    )
    features = cat_features + cont_features + targets

    workflow = nvt.Workflow(features)

    return workflow


def transform_criteo(
    raw_data_path: Union[str, Path],
    output_path: Union[str, Path],
    workflow: Optional[Workflow] = None,
    num_buckets=10000000,
    valid_num_days=1,
    **kwargs,
):
    workflow = workflow or default_criteo_transform(**locals())
    file_list = glob.glob(os.path.join(str(raw_data_path), "day_*[!.gz]"))
    workflow_fit_transform(
        workflow,
        file_list[:-valid_num_days],
        file_list[-valid_num_days],
        str(output_path),
    )

    return output_path



from typing import Optional

import tensorflow as tf
from tensorflow.keras.layers import Layer

from merlin.models.tf.outputs.base import MetricsFn, ModelOutput


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class RegressionOutput(ModelOutput):
    """Regression prediction block

    Parameters
    ----------
    target: str, optional
        The name of the target.
    pre: Optional[Block], optional
        Optional block to transform predictions before computing the regression scores,
        by default None
    post: Optional[Block], optional
        Optional block to transform the regression scores,
        by default None
    name: str, optional
        The name of the task.
    default_loss: Union[str, tf.keras.losses.Loss], optional
        Default loss to use for regression
        by 'mse'
    get_default_metrics: Callable, optional
        A function returning the list of default metrics to set
        if the user does not specify any
        Default metrics to use for regression
    """

    def __init__(
        self,
        target: Optional[str] = None,
        pre: Optional[Layer] = None,
        post: Optional[Layer] = None,
        name: Optional[str] = None,
        default_loss="mse",
        default_metrics_fn: MetricsFn = lambda: (
            tf.keras.metrics.RootMeanSquaredError(name="root_mean_squared_error"),
        ),
        **kwargs,
    ):
        to_call = kwargs.pop("to_call", None)
        super().__init__(
            to_call=to_call or tf.keras.layers.Dense(1, activation="linear"),
            default_loss=default_loss,
            default_metrics_fn=default_metrics_fn,
            target=target,
            pre=pre,
            post=post,
            name=name,
            **kwargs,
        )



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import math
import os
from typing import Dict, Optional

import numpy as np

from merlin.schema import ColumnSchema, Schema, Tags, TagsType
from merlin.schema.io.tensorflow_metadata import TensorflowMetadata


def select_targets(schema: Schema, extra_tags: Optional[TagsType] = None) -> Schema:
    out = schema.select_by_tag(Tags.BINARY_CLASSIFICATION)
    out += schema.select_by_tag(Tags.TARGET)
    out += schema.select_by_tag(Tags.REGRESSION)

    if extra_tags:
        out += schema.select_by_tag(extra_tags)

    return out


def schema_to_tensorflow_metadata_json(schema, path=None):
    json = TensorflowMetadata.from_merlin_schema(schema).to_json()
    if path:
        with open(path, "w") as o:
            o.write(json)
    return json


def tensorflow_metadata_json_to_schema(value):
    if os.path.isfile(value):
        value = open(value).read()
    return TensorflowMetadata.from_json(value).to_merlin_schema()


def create_categorical_column(
    name,
    num_items,
    dtype=np.int32,
    domain_name=None,
    tags=None,
    properties=None,
    min_value_count=None,
    max_value_count=None,
):
    properties = properties or {}
    if not domain_name:
        domain_name = name
    if num_items:
        properties["domain"] = {"name": domain_name, "min": 0, "max": num_items}

    is_list, is_ragged = False, False
    value_count = {}
    if min_value_count is not None:
        value_count["min"] = min_value_count
    if max_value_count is not None:
        value_count["max"] = max_value_count
    if value_count:
        properties["value_count"] = value_count
        is_list = True
        is_ragged = min_value_count != max_value_count

    tags = tags or []
    if Tags.CATEGORICAL not in tags:
        tags.append(Tags.CATEGORICAL)

    return ColumnSchema(
        name=name,
        tags=tags,
        dtype=dtype,
        properties=properties,
        is_list=is_list,
        is_ragged=is_ragged,
    )


def create_continuous_column(
    name,
    dtype=np.float32,
    tags=None,
    properties=None,
    min_value=None,
    max_value=None,
):
    properties = properties or {}
    domain = {}
    if min_value is not None:
        domain["min"] = min_value
    if max_value is not None:
        domain["max"] = max_value
    if domain:
        properties["domain"] = domain

    tags = tags or []
    if Tags.CONTINUOUS not in tags:
        tags.append(Tags.CONTINUOUS)

    return ColumnSchema(name=name, tags=tags, properties=properties, dtype=dtype)


def filter_dict_by_schema(input_dict, schema):
    """Filters out entries from input_dict, returns a dictionary
    where every entry corresponds to a column in the schema"""
    column_names = set(schema.column_names)
    return {k: v for k, v in input_dict.items() if k in column_names}


def categorical_cardinalities(schema) -> Dict[str, int]:
    outputs = {}
    for col in schema:
        if Tags.CATEGORICAL in col.tags:
            domain = col.int_domain
            if domain:
                outputs[col.name] = domain.max + 1

    return outputs


def categorical_domains(schema) -> Dict[str, str]:
    outputs = {}
    for col in schema:
        if Tags.CATEGORICAL in col.tags:
            domain = col.int_domain
            name = col.name
            if domain and domain.name:
                name = domain.name
            outputs[col.name] = name

    return outputs


def get_embedding_sizes_from_schema(
    schema: Schema, multiplier: float = 2.0, ensure_multiple_of_8: bool = False
) -> Dict[str, int]:
    """Provides a heristic (from Google) that suggests the embedding sizes
    as a function (forth root) of categorical features cardinalities, obtained
    from the schema.

    Parameters
    ----------
    schema : Schema
        Featires schema
    multiplier : float, optional
        Multiplier to be applied on the forth root of the cardinality.
        Google recommends multiplier in the [2.0,10.0] range, by default 2.0
    ensure_multiple_of_8 : bool, optional
        If enabled, adjusts the embedding dim to the smallest greater
        number multiple of 8, to ensure best performance with GPU ops, by default False

    Returns
    -------
    Dict[str, int]
        A dict with the feature names and the suggested embedding sizes based
        on the features cardinalities obtained from the schema
    """
    cardinalities = categorical_cardinalities(schema)

    return {
        key: get_embedding_size_from_cardinality(val, multiplier, ensure_multiple_of_8)
        for key, val in cardinalities.items()
    }


def col_is_list(col: ColumnSchema) -> bool:
    return Tags.SEQUENCE in col.tags or Tags.LIST in col.tags or col.is_list


def get_embedding_size_from_cardinality(
    cardinality: int, multiplier: float = 2.0, ensure_multiple_of_8: bool = False
) -> int:
    """Provides a heuristic (from Google) that suggests the embedding
    dimension as a function (forth root) of the feature cardinality.

    Parameters
    ----------
    cardinality : int
        The number of unique values of a categorical feature
    multiplier : float, optional
        Multiplier to be applied on the forth root of the cardinality.
        Google recommends multiplier in the [2.0,10.0] range, by default 2.0
    ensure_multiple_of_8 : bool, optional
        If enabled, adjusts the embedding dim to the smallest greater
        number multiple of 8, to ensure best performance with GPU ops, by default False

    Returns
    -------
    int
        The suggested embedding dimension based on the feature cardinality
    """
    # A rule-of-thumb from Google.
    embedding_size = int(math.ceil(math.pow(cardinality, 0.25) * multiplier))

    if ensure_multiple_of_8:
        embedding_size = int(math.ceil((embedding_size / 8)) * 8)

    return embedding_size


def infer_embedding_dim(
    col_schema: ColumnSchema, multiplier: float = 2.0, ensure_multiple_of_8: bool = True
):
    cardinality = col_schema.int_domain.max + 1

    return get_embedding_size_from_cardinality(
        cardinality, multiplier=multiplier, ensure_multiple_of_8=ensure_multiple_of_8
    )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, Optional, Union

from merlin.models.tf.blocks.retrieval.matrix_factorization import (
    MatrixFactorizationBlock,
    QueryItemIdsEmbeddingsBlock,
)
from merlin.models.tf.core.aggregation import ElementWiseMultiply
from merlin.models.tf.core.base import Block
from merlin.models.tf.core.combinators import ParallelBlock
from merlin.models.tf.models.base import Model
from merlin.models.tf.models.utils import parse_prediction_tasks
from merlin.models.tf.prediction_tasks.base import ParallelPredictionBlock, PredictionTask
from merlin.schema import Schema


def NCFModel(
    schema: Schema,
    embedding_dim: int,
    mlp_block: Block,
    prediction_tasks: Optional[
        Union[PredictionTask, List[PredictionTask], ParallelPredictionBlock]
    ] = None,
    embeddings_l2_reg: float = 0.0,
    **kwargs,
) -> Model:
    """NCF-model architecture.

    Example Usage::
        ncf = NCFModel(schema, embedding_dim=64, mlp_block=MLPBlock([256, 64]))
        ncf.compile(optimizer="adam")
        ncf.fit(train_data, epochs=10)

    References
    ----------
    [1] Xiangnan,
        He, et al. "Neural Collaborative Filtering."
        arXiv:1708.05031 (2017).
    [2] Steffen, Rendle et al.
        "Neural Collaborative Filtering vs. Matrix Factorization Revisited."
        arXiv:2005.09683(2020)

    Notes
    -----
    We note that  Rendle et al. [2] showed that properly initialized MF
    significantly outperforms the NCF model.

    Parameters
    ----------
    schema : Schema
        The `Schema` with the input features
    embedding_dim : int
        Dimension of the embeddings
    mlp_block : MLPBlock
        Stack of MLP layers to learn non-linear interactions from data.
    prediction_tasks: optional
        The prediction tasks to be used, by default this will be inferred from the Schema.
    embeddings_l2_reg: float = 0.0
        Factor for L2 regularization of the embeddings vectors (from the current batch only)
    Returns
    -------
    Model

    """

    mlp_branch = QueryItemIdsEmbeddingsBlock(
        schema, dim=embedding_dim, embeddings_l2_reg=embeddings_l2_reg
    ).connect(mlp_block)
    mf_branch = MatrixFactorizationBlock(
        schema,
        dim=embedding_dim,
        aggregation=ElementWiseMultiply(),
        embeddings_l2_reg=embeddings_l2_reg,
        **kwargs,
    )

    ncf = ParallelBlock({"mf": mf_branch, "mlp": mlp_branch}, aggregation="concat")

    prediction_tasks = parse_prediction_tasks(schema, prediction_tasks)
    model = Model(ncf, prediction_tasks)

    return model



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import os
import warnings
from pathlib import Path
from typing import Dict, List, Optional, Union

import distributed
import numpy as np
import xgboost as xgb

from merlin.core.utils import global_dask_client
from merlin.io import Dataset
from merlin.models.utils.schema_utils import (
    schema_to_tensorflow_metadata_json,
    tensorflow_metadata_json_to_schema,
)
from merlin.schema import Schema, Tags


class XGBoost:
    """Create an XGBoost model from a merlin dataset.
    The class adapts an XGBoost model to work with the high level merlin-models API.

    Example usage::

        # get the movielens dataset
        from merlin.datasets.entertainment import get_movielens

        train, valid = get_movielens(variant="ml-1m")

        # Train an XGBoost model
        from merlin.core.utils import Distributed
        from merlin.models.xgb import XGBoost

        with Distributed():
            model = XGBoost(train.schema, objective="binary:logistic")
            model.fit(train)
            metrics = model.evaluate(valid)
    """

    def __init__(
        self,
        schema: Schema,
        *,
        target_columns: Optional[Union[str, list]] = None,
        qid_column: Optional[str] = None,
        objective: str = "reg:squarederror",
        booster: Optional[xgb.Booster] = None,
        **params,
    ):
        """
        Parameters
        ----------
        schema : merlin.schema.Schema
            The schema of the data that will be used to train and evaluate the model.
        target_columns : Optional[Union[list, str]]
            The target columns to use. If provided, will be used as the label(s).
            Otherwise the targets are automatically inferred from the objective and column tags.
        qid_column : Optional[str]
            For ranking objectives. The query ID column. If not provided will use
            the user ID (tagged with merlin.schema.Tags.USER_ID) column.
        objective : str
            The XGBoost objective to use. List of XGBoost objective functions:
            https://xgboost.readthedocs.io/en/stable/gpu/index.html#objective-functions
        **params
            The parameters to use for the XGBoost train method
        """
        self.params = {**params, "objective": objective}

        target_tag = get_target_tag(objective)
        if isinstance(target_columns, str):
            target_columns = [target_columns]
        self.target_columns = target_columns or get_targets(schema, target_tag)
        self.feature_columns = get_features(schema, self.target_columns)

        if objective.startswith("rank") and qid_column is None:
            qid_column = schema.select_by_tag(Tags.USER_ID).column_names[0]
        self.qid_column = qid_column
        self.evals_result = {}
        self.booster = booster
        self.schema = schema

    @property
    def dask_client(self) -> Optional[distributed.Client]:
        return global_dask_client()

    def fit(
        self,
        train: Dataset,
        *,
        evals=None,
        use_quantile=True,
        **train_kwargs,
    ) -> xgb.Booster:
        """Trains the XGBoost Model.

        Will use the columns tagged with the target_type passed to the
        constructor as the labels.  And all other non-list columns as
        input features.

        Parameters
        ----------
        train : merlin.io.Dataset
            The training dataset to use to fit the model.
            We will use the column(s) tagged with merlin.schema.Tags.TARGET that match the
            objective as the label(s).
        evals : List[Tuple[Dataset, str]]
            List of tuples of datasets to watch
        use_quantile : bool
            This param is only relevant when using GPU.  (with
            tree_method="gpu_hist"). If set to False, will use a
            `DaskDMatrix`, instead of the default
            `DaskDeviceQuantileDMatrix`, which is preferred for GPU training.
        **train_kwargs
            Additional keyword arguments passed to the xgboost.train function

        Returns
        -------
        xgb.Booster

        Raises
        ------
        ValueError
           If objective is not supported. Or if the target columns cannot be found.
        """
        X, y, qid = dataset_to_xy(
            train,
            self.feature_columns,
            self.target_columns,
            self.qid_column,
        )

        dmatrix_cls = xgb.dask.DaskDMatrix
        if self.params.get("tree_method") == "gpu_hist" and use_quantile:
            # `DaskDeviceQuantileDMatrix` is a data type specialized
            # for the `gpu_hist` tree method that reduces memory overhead.
            # When training on GPU pipeline, it's preferred over `DaskDMatrix`.
            dmatrix_cls = xgb.dask.DaskDeviceQuantileDMatrix

        dtrain = dmatrix_cls(self.dask_client, X, label=y, qid=qid)
        watchlist = []

        if evals is None:
            evals = [(train, "train")]

        for _eval in evals:
            assert len(_eval) == 2
            dataset, name = _eval
            if dataset == train:
                watchlist.append((dtrain, name))
                continue
            assert isinstance(dataset, Dataset)
            X, y, qid = dataset_to_xy(
                dataset,
                self.feature_columns,
                self.target_columns,
                self.qid_column,
            )
            d_eval = dmatrix_cls(self.dask_client, X, label=y, qid=qid)
            watchlist.append((d_eval, name))

        train_res = xgb.dask.train(
            self.dask_client, self.params, dtrain, evals=watchlist, **train_kwargs
        )
        self.booster: xgb.Booster = train_res["booster"]
        self.evals_result = train_res["history"]

        return self.booster

    def evaluate(self, dataset: Dataset, **predict_kwargs) -> Dict[str, float]:
        """Evaluates the model on the dataset provided.

        Parameters
        ----------
        dataset : merlin.io.Dataset
            The dataset used to evaluate the model.
        **predict_kwargs
            The keyword parameters passed to the predict function

        Returns
        -------
        Dict[str, float]
            Dictionary of metrics of the form {metric_name: value}.
        """
        if self.booster is None:
            raise ValueError("The fit method must be called before evaluate.")

        X, y, qid = dataset_to_xy(
            dataset, self.feature_columns, self.target_columns, self.qid_column
        )

        # convert to DMatrix
        # (eval doesn't have dask support currently)
        if qid is not None:
            qid = qid.compute()
        eval_data = xgb.DMatrix(X.compute(), label=y.compute(), qid=qid)

        metrics_str = self.booster.eval(eval_data)
        metrics = {}
        for metric in metrics_str.split("\t")[1:]:
            metric_name, metric_value = metric.split(":")
            metrics[metric_name[len("eval-") :]] = float(metric_value)

        return metrics

    def predict(self, dataset: Dataset, **predict_kwargs) -> np.ndarray:
        """Generate predictions from the dataset.

        Parameters
        ----------
        dataset : merlin.io.Dataset
            The dataset to use for predictions
        **predict_kwargs
            keyword arguments passed to the xgboost.core.Booster.predict method

        Returns
        -------
        numpy.ndarray
            The predicions data
        """
        if self.booster is None:
            raise ValueError("The fit method must be called before predict.")

        X, _, qid = dataset_to_xy(dataset, self.feature_columns, [], self.qid_column)
        data = xgb.dask.DaskDMatrix(self.dask_client, X, qid=qid)
        preds = xgb.dask.predict(self.dask_client, self.booster, data, **predict_kwargs).compute()

        return preds

    def save(self, path: Union[str, os.PathLike]) -> None:
        """Save the model to a directory.

        Parameters
        ----------
        path : Union[str, os.PathLike]
            Directory where the model will be saved.
        """
        export_dir = Path(path)
        export_dir.mkdir(parents=True)
        self.booster.save_model(export_dir / "model.json")
        schema_to_tensorflow_metadata_json(self.schema, export_dir / "schema.json")
        with open(export_dir / "params.json", "w") as f:
            json.dump(self.params, f, indent=4)
        with open(export_dir / "config.json", "w") as f:
            json.dump(
                dict(qid_column=self.qid_column, target_columns=self.target_columns), f, indent=4
            )

    @classmethod
    def load(cls, path: Union[str, os.PathLike]) -> "XGBoost":
        """Load the model from a directory where a model has been saved.

        Parameters
        ----------
        path : Union[str, os.PathLike]
            Path where a Merlin XGBoost model has been saved.

        Returns
        -------
        XGBoost model instance
        """
        load_dir = Path(path)
        booster = xgb.Booster()
        booster.load_model(load_dir / "model.json")
        schema = tensorflow_metadata_json_to_schema(load_dir / "schema.json")
        with open(load_dir / "params.json", "r") as f:
            params = json.load(f)
        with open(load_dir / "config.json", "r") as f:
            config = json.load(f)
        return cls(
            schema,
            target_columns=config.get("target_columns"),
            qid_column=config.get("qid_column"),
            booster=booster,
            **params,
        )


OBJECTIVES = {
    "binary:logistic": Tags.BINARY_CLASSIFICATION,
    "reg:logistic": Tags.REGRESSION,
    "reg:squarederror": Tags.REGRESSION,
    "rank:pairwise": Tags.TARGET,
    "rank:ndcg": Tags.TARGET,
    "rank:map": Tags.TARGET,
}


def get_target_tag(objective: str) -> Tags:
    """Get the target tag from the specified objective"""
    try:
        return OBJECTIVES[objective]
    except KeyError as exc:
        target_options_str = str(list(OBJECTIVES.keys()))
        raise ValueError(f"Objective not supported. Must be one of: {target_options_str}") from exc


def get_targets(schema: Schema, target_tag: Tags) -> List[str]:
    """Find target columns from dataset or specified target_column"""
    targets = schema.select_by_tag(Tags.TARGET).select_by_tag(target_tag)

    if len(targets) >= 1:
        return targets.column_names
    raise ValueError(
        f"No target columns in the dataset schema with tags TARGET and {target_tag.name}"
    )


def get_features(schema: Schema, target_columns: List[str]):
    """Find feature columns from schema. Returns all non-list column names from the schema
    that are not tagged as targets."""
    all_target_columns = schema.select_by_tag(Tags.TARGET).column_names + target_columns

    # Ignore list-like columns from schema
    list_column_names = [
        col_name for col_name, col_schema in schema.column_schemas.items() if col_schema.is_list
    ]

    if list_column_names:
        warnings.warn(f"Ignoring list columns as inputs to XGBoost model: {list_column_names}.")

    feature_columns = schema.excluding_by_name(
        set(list_column_names + all_target_columns)
    ).column_names
    if len(feature_columns) == 0:
        raise ValueError("No feature columns found in schema.")

    return feature_columns


def dataset_to_xy(
    dataset: Dataset,
    feature_columns: List[str],
    target_columns: List[str],
    qid_column: Optional[str],
):
    """Convert Merlin Dataset to XGBoost DMatrix"""
    df = dataset.to_ddf()

    qid = None
    if qid_column:
        df = df.sort_values(qid_column)
        qid = df[qid_column]

    X = df[feature_columns]
    y = df[target_columns]

    return X, y, qid



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np

from merlin.datasets.synthetic import generate_data
from merlin.models.lightfm import LightFM
from merlin.schema import Tags


def test_warp():
    np.random.seed(0)
    train, valid = generate_data("music-streaming", 100, (0.95, 0.05))
    train.schema = train.schema.remove_by_tag(Tags.TARGET)
    valid.schema = valid.schema.remove_by_tag(Tags.TARGET)

    model = LightFM(learning_rate=0.05, loss="warp", epochs=10)
    model.fit(train)

    model.predict(train)

    metrics = model.evaluate(valid, k=10)
    assert metrics["auc"] > 0.01



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
import pandas as pd
import pytest
import tensorflow as tf
from tensorflow.test import TestCase

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.transforms.features import BroadcastToSequence, ContinuousPowers
from merlin.models.tf.utils import testing_utils
from merlin.models.utils.schema_utils import create_categorical_column
from merlin.schema import ColumnSchema, Schema, Tags


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_categorical_encoding_as_pre(ecommerce_data: Dataset, run_eagerly):
    schema = ecommerce_data.schema.select_by_name(names=["user_categories", "item_category"])
    body = mm.ParallelBlock(
        mm.TabularBlock.from_schema(schema=schema, pre=mm.CategoryEncoding(schema)),
        is_input=True,
    ).connect(mm.MLPBlock([32]))
    model = mm.Model(body, mm.BinaryClassificationTask("click"))

    testing_utils.model_test(model, ecommerce_data, run_eagerly=run_eagerly)


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_categorical_encoding_in_model(ecommerce_data: Dataset, run_eagerly):
    schema = ecommerce_data.schema.select_by_name(names=["user_categories", "item_category"])
    branches = {
        "one_hot": mm.CategoryEncoding(schema, is_input=True),
        "features": mm.InputBlock(ecommerce_data.schema),
    }
    body = mm.ParallelBlock(branches, is_input=True).connect(mm.MLPBlock([32]))
    model = mm.Model(body, mm.BinaryClassificationTask("click"))

    testing_utils.model_test(model, ecommerce_data, run_eagerly=run_eagerly)


def test_continuous_powers():
    NUM_ROWS = 100

    inputs = {
        "cont_feat_1": tf.random.uniform((NUM_ROWS,)),
        "cont_feat_2": tf.random.uniform((NUM_ROWS,)),
    }

    powers = ContinuousPowers()

    outputs = powers(inputs)

    assert len(outputs) == len(inputs) * 3
    for key in inputs:
        assert key in outputs
        assert key + "_sqrt" in outputs
        assert key + "_pow" in outputs

        tf.assert_equal(tf.sqrt(inputs[key]), outputs[key + "_sqrt"])
        tf.assert_equal(tf.pow(inputs[key], 2), outputs[key + "_pow"])


def test_hashedcross_int():
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=3),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=3),
        ]
    )
    inputs = {}
    inputs["cat1"] = tf.constant(["A", "B"])
    inputs["cat2"] = tf.constant([101, 102])
    hashed_cross_op = mm.HashedCross(schema=schema, num_bins=10, output_mode="int")
    outputs = hashed_cross_op(inputs)
    output_name, output_value = outputs.popitem()

    assert output_name == "cross_cat1_cat2"
    assert output_value.shape.as_list() == [2, 1]
    test_case.assertAllClose(output_value, [[1], [6]])


def test_hashedcross_1d():
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=20),
        ]
    )
    inputs = {}
    inputs["cat1"] = tf.constant(["A", "B", "A", "B", "A"])
    inputs["cat2"] = tf.constant([101, 101, 101, 102, 102])
    hashed_cross_op = mm.HashedCross(schema=schema, num_bins=10, output_mode="int")
    outputs = hashed_cross_op(inputs)
    _, output_value = outputs.popitem()

    assert output_value.shape.as_list() == [5, 1]
    test_case.assertAllClose(output_value, [[1], [4], [1], [6], [3]])


def test_hashedcross_2d():
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=20),
        ]
    )
    inputs = {}
    inputs["cat1"] = tf.constant([["A"], ["B"], ["A"], ["B"], ["A"]])
    inputs["cat2"] = tf.constant([[101], [101], [101], [102], [102]])
    hashed_cross_op = mm.HashedCross(schema=schema, num_bins=10, output_mode="int")
    outputs = hashed_cross_op(inputs)
    _, output_value = outputs.popitem()

    assert output_value.shape.as_list() == [5, 1]
    test_case.assertAllClose(output_value, [[1], [4], [1], [6], [3]])


def test_hashedcross_output_shape():
    schema = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=20),
        ]
    )
    inputs_shape = {}
    inputs_shape["cat1"] = tf.constant([["A"], ["B"], ["A"], ["B"], ["A"]]).shape
    inputs_shape["cat2"] = tf.constant([[101], [101], [101], [102], [102]]).shape
    hashed_cross = mm.HashedCross(schema=schema, num_bins=10, output_mode="int")
    outputs = hashed_cross.compute_output_shape(inputs_shape)
    _, output_shape = outputs.popitem()

    assert output_shape == [5, 1]


def test_hashedcross_output_shape_one_hot():
    schema = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=20),
        ]
    )
    inputs_shape = {}
    inputs_shape["cat1"] = tf.constant([["A"], ["B"], ["A"], ["B"], ["A"]]).shape
    inputs_shape["cat2"] = tf.constant([[101], [101], [101], [102], [102]]).shape
    output_name = "cross_out"
    hashed_cross = mm.HashedCross(
        schema=schema, num_bins=10, output_mode="one_hot", output_name=output_name
    )
    outputs = hashed_cross.compute_output_shape(inputs_shape)
    _output_name, output_shape = outputs.popitem()

    assert output_shape == [5, 10]
    assert _output_name == output_name


def test_hashedcross_less_bins():
    schema = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=20),
        ]
    )
    inputs = {}
    inputs["cat1"] = tf.constant([["A"], ["B"], ["C"], ["D"], ["A"], ["B"], ["A"]])
    inputs["cat2"] = tf.constant([[101], [102], [101], [101], [101], [102], [103]])
    hashed_cross_op = mm.HashedCross(schema=schema, num_bins=4, output_mode="one_hot", sparse=True)
    outputs = hashed_cross_op(inputs)
    _, output_value = outputs.popitem()
    output_value = tf.sparse.to_dense(output_value)

    assert output_value.shape.as_list() == [7, 4]


def test_hashedcross_output_mode():
    schema = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=20),
        ]
    )
    inputs = {}
    inputs["cat1"] = tf.constant([["A"], ["B"], ["C"], ["D"], ["A"], ["B"], ["A"]])
    inputs["cat2"] = tf.constant([[101], [102], [101], [101], [101], [102], [103]])

    hashed_cross_op = mm.HashedCross(schema=schema, num_bins=4, output_mode="one_hot", sparse=True)
    outputs = hashed_cross_op(inputs)
    _, output_value = outputs.popitem()
    assert isinstance(output_value, tf.SparseTensor) is True
    assert output_value.shape.as_list() == [7, 4]

    hashed_cross_op = mm.HashedCross(schema=schema, num_bins=4, output_mode="one_hot", sparse=False)
    outputs = hashed_cross_op(inputs)
    _, output_value = outputs.popitem()
    assert isinstance(output_value, tf.Tensor) is True
    assert output_value.shape.as_list() == [7, 4]


def test_hashedcross_onehot_output():
    test_case = TestCase()

    schema = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=20),
        ]
    )
    inputs = {}
    inputs["cat1"] = tf.constant([["A"], ["B"], ["A"], ["B"], ["A"]])
    inputs["cat2"] = tf.constant([[101], [101], [101], [102], [102]])
    hashed_cross_op = mm.HashedCross(schema=schema, num_bins=5, output_mode="one_hot", sparse=True)
    outputs = hashed_cross_op(inputs)
    _, output_value = outputs.popitem()
    output_value = tf.sparse.to_dense(output_value)

    assert output_value.shape.as_list() == [5, 5]
    test_case.assertAllClose(
        output_value,
        [
            [0, 1, 0, 0, 0],
            [0, 0, 0, 0, 1],
            [0, 1, 0, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 0, 1, 0],
        ],
    )


def test_hashedcross_single_input_fails():
    test_case = TestCase()
    schema = Schema([create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20)])
    with test_case.assertRaisesRegex(ValueError, "at least two features"):
        mm.HashedCross(num_bins=10, schema=schema, output_mode="int")([tf.constant(1)])


def test_hashedcross_from_config():
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=20),
        ]
    )
    inputs = {}
    inputs["cat1"] = tf.constant([["A"], ["B"], ["A"], ["B"], ["A"]])
    inputs["cat2"] = tf.constant([[101], [101], [101], [102], [102]])
    hashed_cross_op = mm.HashedCross(schema=schema, num_bins=5, output_mode="one_hot", sparse=False)
    cloned_hashed_cross_op = mm.HashedCross.from_config(hashed_cross_op.get_config())
    original_outputs = hashed_cross_op(inputs)
    cloned_outputs = cloned_hashed_cross_op(inputs)
    _, original_output_value = original_outputs.popitem()
    _, cloned_output_value = cloned_outputs.popitem()

    test_case.assertAllEqual(cloned_output_value, original_output_value)


def test_hashedcrosses_in_parallelblock():
    test_case = TestCase()

    schema_0 = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=20),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=20),
        ]
    )
    schema_1 = Schema(
        [
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=2),
            create_categorical_column("cat3", tags=[Tags.CATEGORICAL], num_items=3),
        ]
    )
    inputs = {}
    inputs["cat1"] = tf.constant([["A"], ["B"], ["A"], ["B"], ["A"]])
    inputs["cat2"] = tf.constant([[101], [101], [101], [102], [102]])
    inputs["cat3"] = tf.constant([[1], [0], [1], [2], [2]])
    hashed_cross_0 = mm.HashedCross(
        schema=schema_0, num_bins=5, output_mode="one_hot", sparse=True, output_name="cross_0"
    )
    hashed_cross_1 = mm.HashedCross(
        schema=schema_1, num_bins=10, output_mode="one_hot", sparse=True, output_name="cross_1"
    )
    hashed_crosses = mm.ParallelBlock([hashed_cross_0, hashed_cross_1])
    outputs = hashed_crosses(inputs)
    output_value_0 = outputs["cross_0"]
    output_value_0 = tf.sparse.to_dense(output_value_0)

    assert output_value_0.shape.as_list() == [5, 5]
    test_case.assertAllClose(
        output_value_0,
        [
            [0, 1, 0, 0, 0],
            [0, 0, 0, 0, 1],
            [0, 1, 0, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 0, 1, 0],
        ],
    )
    output_value_1 = outputs["cross_1"]
    output_value_1 = tf.sparse.to_dense(output_value_1)

    assert output_value_1.shape.as_list() == [5, 10]
    test_case.assertAllClose(
        output_value_1,
        [
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
        ],
    )


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_hashedcross_as_pre(ecommerce_data: Dataset, run_eagerly):
    cross_schema = ecommerce_data.schema.select_by_name(names=["user_categories", "item_category"])
    body = mm.ParallelBlock(
        mm.TabularBlock.from_schema(
            schema=cross_schema, pre=mm.HashedCross(cross_schema, num_bins=1000)
        ),
        is_input=True,
    ).connect(mm.MLPBlock([64]))
    model = mm.Model(body, mm.BinaryClassificationTask("click"))

    testing_utils.model_test(model, ecommerce_data, run_eagerly=run_eagerly)


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_hashedcross_in_model(ecommerce_data: Dataset, run_eagerly):
    cross_schema = ecommerce_data.schema.select_by_name(names=["user_categories", "item_category"])
    branches = {
        "cross_product": mm.HashedCross(cross_schema, num_bins=1000, is_input=True),
        "features": mm.InputBlock(ecommerce_data.schema),
    }
    body = mm.ParallelBlock(branches, is_input=True).connect(mm.MLPBlock([64]))
    model = mm.Model(body, mm.BinaryClassificationTask("click"))

    model.compile(optimizer="adam", run_eagerly=run_eagerly)
    testing_utils.model_test(model, ecommerce_data)


def test_category_encoding_different_input_different_output():
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("dense_feature", tags=[Tags.CATEGORICAL], num_items=4),
            create_categorical_column("sparse_feature", tags=[Tags.CATEGORICAL], num_items=5),
        ]
    )
    inputs = {}
    inputs["dense_feature"] = tf.constant([[1, 2, 3], [3, 3, 0]])
    inputs["sparse_feature"] = tf.sparse.from_dense(
        np.array([[1, 2, 3, 0], [0, 3, 1, 0]], dtype=np.int64)
    )

    # 1. Sparse output
    category_encoding = mm.CategoryEncoding(
        schema=schema,
        output_mode="count",
        sparse=True,
    )
    outputs = category_encoding(inputs)

    # The expected output["dense_feature"] should be (X for missing value):
    # [[X, 1, 1, 1]
    #  [1, X, X, 2]]
    expected_indices_1 = [[0, 1], [0, 2], [0, 3], [1, 0], [1, 3]]
    expected_values_1 = [1, 1, 1, 1, 2]
    test_case.assertAllEqual(expected_values_1, outputs["dense_feature"].values)
    test_case.assertAllEqual(expected_indices_1, outputs["dense_feature"].indices)

    expected_indices_2 = [[0, 1], [0, 2], [0, 3], [1, 1], [1, 3]]
    expected_values_2 = [1, 1, 1, 1, 1]
    test_case.assertAllEqual(expected_values_2, outputs["sparse_feature"].values)
    test_case.assertAllEqual(expected_indices_2, outputs["sparse_feature"].indices)

    # 2. Dense output
    category_encoding = mm.CategoryEncoding(
        schema=schema,
        output_mode="count",
        sparse=False,
    )
    expected_1 = [[0, 1, 1, 1, 0], [1, 0, 0, 2, 0]]
    expected_2 = [[0, 1, 1, 1, 0, 0], [0, 1, 0, 1, 0, 0]]
    outputs = category_encoding(inputs)
    test_case.assertAllEqual(expected_1, outputs["dense_feature"])
    test_case.assertAllEqual(expected_2, outputs["sparse_feature"])


def test_category_encoding_invalid_input():
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("ragged_feature", tags=[Tags.CATEGORICAL], num_items=5),
        ]
    )
    inputs = {}
    inputs["ragged_feature"] = tf.ragged.constant([[1, 2, 3], [3, 1], []])
    category_encoding = mm.CategoryEncoding(
        schema=schema,
        output_mode="count",
        sparse=False,
    )
    with test_case.assertRaisesRegex(ValueError, "inputs should not contain a RaggedTensor"):
        category_encoding(inputs)


@pytest.mark.parametrize("input", [np.array([[1, 2, 3, 4], [4, 3, 1, 4]])])
@pytest.mark.parametrize("weight", [np.array([[0.1, 0.2, 0.3, 0.4], [0.2, 0.1, 0.4, 0.3]])])
def test_category_encoding_weightd_count_dense(input, weight):
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("feature", tags=[Tags.CATEGORICAL], num_items=5),
        ]
    )

    expected_output = [[0, 0.1, 0.2, 0.3, 0.4, 0], [0, 0.4, 0, 0.1, 0.5, 0]]
    # pyformat: enable
    expected_output_shape = [2, 6]

    category_encoding = mm.CategoryEncoding(
        schema=schema, output_mode="count", count_weights=weight
    )

    inputs = {}
    inputs["feature"] = input
    outputs = category_encoding(inputs)
    test_case.assertAllEqual(expected_output_shape, outputs["feature"].shape.as_list())
    test_case.assertAllClose(expected_output, outputs["feature"])


@pytest.mark.parametrize("input", [tf.sparse.from_dense(np.array([[1, 2, 3, 4], [4, 3, 1, 4]]))])
@pytest.mark.parametrize(
    "weight", [tf.sparse.from_dense(np.array([[0.1, 0.2, 0.3, 0.4], [0.2, 0.1, 0.4, 0.3]]))]
)
def test_category_encoding_weightd_count_sparse(input, weight):
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("feature", tags=[Tags.CATEGORICAL], num_items=5),
        ]
    )

    expected_output = [[0, 0.1, 0.2, 0.3, 0.4, 0], [0, 0.4, 0, 0.1, 0.5, 0]]
    # pyformat: enable
    expected_output_shape = [2, 6]

    category_encoding = mm.CategoryEncoding(
        schema=schema, output_mode="count", count_weights=weight
    )

    inputs = {}
    inputs["feature"] = input
    outputs = category_encoding(inputs)
    test_case.assertAllEqual(expected_output_shape, outputs["feature"].shape.as_list())
    test_case.assertAllClose(expected_output, outputs["feature"])


@pytest.mark.parametrize("input", [tf.sparse.from_dense(np.array([[1, 2, 3, 4], [4, 3, 1, 4]]))])
@pytest.mark.parametrize("weight", [np.array([[0.1, 0.2, 0.3, 0.4], [0.2, 0.1, 0.4, 0.3]])])
def test_category_encoding_weightd_count_not_match(input, weight):
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("feature", tags=[Tags.CATEGORICAL], num_items=5),
        ]
    )
    category_encoding = mm.CategoryEncoding(
        schema=schema, output_mode="count", count_weights=weight
    )
    inputs = {}
    inputs["feature"] = input
    with test_case.assertRaisesRegex(
        ValueError, "Argument `weights` must be a SparseTensor if `values` is a SparseTensor"
    ):
        category_encoding(inputs)


@pytest.mark.parametrize(
    "input",
    [
        tf.convert_to_tensor([[1, 2, 3, 0], [0, 3, 1, 0]]),
        tf.sparse.from_dense(np.array([[1, 2, 3, 0], [0, 3, 1, 0]])),
    ],
)
def test_category_encoding_multi_hot_2d_input(input):
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("feature", tags=[Tags.CATEGORICAL], num_items=5),
        ]
    )

    if isinstance(input, tf.SparseTensor):
        expected_output = [[0, 1, 1, 1, 0, 0], [0, 1, 0, 1, 0, 0]]
    else:
        # Dense tensors with 0 will have it included in the multi-hot output
        expected_output = [[1, 1, 1, 1, 0, 0], [1, 1, 0, 1, 0, 0]]
    # pyformat: enable
    expected_output_shape = [2, 6]

    category_encoding = mm.CategoryEncoding(schema=schema, output_mode="multi_hot")

    inputs = {}
    inputs["feature"] = input
    outputs = category_encoding(inputs)
    test_case.assertAllEqual(expected_output_shape, outputs["feature"].shape.as_list())
    test_case.assertAllClose(expected_output, outputs["feature"])


@pytest.mark.parametrize(
    "input",
    [
        tf.convert_to_tensor([1, 2, 0]),
        tf.sparse.from_dense([1, 2, 0]),
        tf.convert_to_tensor([[1], [2], [0]]),
        tf.sparse.from_dense([[1], [2], [0]]),
    ],
)
def test_category_encoding_multi_hot_single_value(input):
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("feature", tags=[Tags.CATEGORICAL], num_items=5),
        ]
    )

    if isinstance(input, tf.SparseTensor):
        expected_output = [[0, 1, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 0]]
    else:
        # Dense tensors with 0 will have it included in the multi-hot output
        expected_output = [[0, 1, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0], [1, 0, 0, 0, 0, 0]]
    # pyformat: enable
    expected_output_shape = [3, 6]

    category_encoding = mm.CategoryEncoding(schema=schema, output_mode="multi_hot")

    inputs = {}
    inputs["feature"] = input
    outputs = category_encoding(inputs)
    test_case.assertAllEqual(expected_output_shape, outputs["feature"].shape.as_list())
    test_case.assertAllClose(expected_output, outputs["feature"])


@pytest.mark.parametrize(
    "input",
    [
        tf.convert_to_tensor([1, 2, 3, 0]),
        tf.sparse.from_dense(np.array([1, 2, 3, 0])),
        tf.convert_to_tensor([[1], [2], [3], [0]]),
        tf.sparse.from_dense(np.array([[1], [2], [3], [0]])),
    ],
)
def test_category_encoding_one_hot(input):
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("feature", tags=[Tags.CATEGORICAL], num_items=5),
        ]
    )

    if isinstance(input, tf.SparseTensor):
        expected_output = [
            [0, 1, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0],
            [0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0],
        ]
    else:
        # Dense tensors with 0 will have it included in the multi-hot output
        expected_output = [
            [0, 1, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0],
            [0, 0, 0, 1, 0, 0],
            [1, 0, 0, 0, 0, 0],
        ]
    # pyformat: enable
    expected_output_shape = [4, 6]

    category_encoding = mm.CategoryEncoding(schema=schema, output_mode="one_hot")

    inputs = {}
    inputs["feature"] = input
    outputs = category_encoding(inputs)
    test_case.assertAllEqual(expected_output_shape, outputs["feature"].shape.as_list())
    test_case.assertAllClose(expected_output, outputs["feature"])


@pytest.mark.parametrize(
    "input",
    [tf.convert_to_tensor([[1, 2], [2, 0]]), tf.sparse.from_dense(np.array([[1, 2], [2, 0]]))],
)
def test_category_encoding_one_hot_2D_input_should_raise(input):
    test_case = TestCase()
    schema = Schema([create_categorical_column("feature", tags=[Tags.CATEGORICAL], num_items=5)])
    category_encoding = mm.CategoryEncoding(schema=schema, output_mode="one_hot")
    inputs = {}
    inputs["feature"] = input
    with test_case.assertRaisesRegex(
        ValueError, r"One-hot accepts input tensors that are squeezable to 1D"
    ):
        category_encoding(inputs)


@pytest.mark.parametrize(
    "input",
    [
        tf.convert_to_tensor([[[1], [2]], [[2], [0]]]),
        tf.sparse.from_dense(np.array([[[1], [2]], [[2], [0]]])),
    ],
)
def test_category_encoding_should_raise_if_input_3D(input):
    test_case = TestCase()
    schema = Schema(
        [
            create_categorical_column("feature", tags=[Tags.CATEGORICAL], num_items=5),
        ]
    )
    category_encoding = mm.CategoryEncoding(schema=schema, output_mode="multi_hot")
    inputs = {}
    inputs["feature"] = input
    with test_case.assertRaisesRegex(
        Exception, r"`CategoryEncoding` only accepts 1D or 2D-shaped inputs"
    ):
        category_encoding(inputs)


def test_hashedcrossall():
    schema = Schema(
        [
            # num_items: 0, 1, 2 thus cardinality = 3
            create_categorical_column("cat1", tags=[Tags.CATEGORICAL], num_items=2),
            create_categorical_column("cat2", tags=[Tags.CATEGORICAL], num_items=2),
            create_categorical_column("cat3", tags=[Tags.CATEGORICAL], num_items=2),
            create_categorical_column("cat4", tags=[Tags.CATEGORICAL], num_items=3),
            create_categorical_column("cat5", tags=[Tags.CATEGORICAL], num_items=3),
            create_categorical_column("cat6", tags=[Tags.CATEGORICAL], num_items=3),
        ]
    )
    inputs = {}
    inputs["cat1"] = tf.constant([["A"], ["B"], ["A"], ["B"], ["A"]])
    inputs["cat2"] = tf.constant([[101], [101], [101], [102], [102]])
    inputs["cat3"] = tf.constant([[1], [0], [1], [2], [2]])
    inputs["cat4"] = tf.constant([[1], [0], [1], [3], [2]])
    inputs["cat5"] = tf.constant([[1], [0], [1], [3], [2]])
    inputs["cat6"] = tf.constant([[1], [0], [1], [3], [2]])

    hashed_cross_all = mm.HashedCrossAll(
        schema=schema,
        infer_num_bins=True,
        output_mode="one_hot",
        sparse=True,
        max_num_bins=25,
        max_level=3,
        ignore_combinations=[["cat3", "cat4", "cat5"], ["cat1", "cat2"]],
    )

    outputs = hashed_cross_all(inputs)
    assert len(outputs) == 33

    output_value_0 = outputs["cross_cat1_cat3"]
    assert output_value_0.shape.as_list() == [5, 9]

    output_value_1 = outputs["cross_cat1_cat3_cat6"]
    assert output_value_1.shape.as_list() == [5, 25]


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_hashedcrossall_in_model(ecommerce_data: Dataset, run_eagerly):
    cross_schema = ecommerce_data.schema.select_by_name(
        names=["user_categories", "item_category", "item_brand"]
    )
    branches = {
        "cross_product": mm.HashedCrossAll(cross_schema, max_num_bins=1000, infer_num_bins=True),
        "features": mm.InputBlock(ecommerce_data.schema),
    }
    body = mm.ParallelBlock(branches, is_input=True).connect(mm.MLPBlock([64]))
    model = mm.Model(body, mm.BinaryClassificationTask("click"))

    testing_utils.model_test(model, ecommerce_data, run_eagerly=run_eagerly)


class TestBroadcastToSequence(tf.test.TestCase):
    def test_only_sequential(self):
        context_schema = Schema()
        sequence_schema = Schema([ColumnSchema("s1", tags=[Tags.SEQUENCE])])
        layer = BroadcastToSequence(context_schema, sequence_schema)
        batch_size = 2
        sequence_length = 6
        inputs = {
            "s1": tf.random.uniform((batch_size, sequence_length, 1)),
        }
        outputs = layer(inputs)
        assert list(outputs["s1"].shape) == [batch_size, sequence_length, 1]

    def test_broadcast(self):
        context_schema = Schema([ColumnSchema("c1")])
        sequence_schema = Schema([ColumnSchema("s1")])
        layer = BroadcastToSequence(context_schema, sequence_schema)
        inputs = {
            "c1": tf.constant([[1], [2]]),
            "s1": tf.constant([[[1, 2], [3, 4], [5, 6]], [[6, 3], [2, 3], [7, 3]]]),
        }
        outputs = layer(inputs)
        self.assertAllEqual(
            outputs["c1"],
            tf.constant(
                [
                    [
                        [1],
                        [1],
                        [1],
                    ],
                    [
                        [2],
                        [2],
                        [2],
                    ],
                ]
            ),
        )
        self.assertAllEqual(inputs["s1"], outputs["s1"])

    def test_different_sequence_lengths(self):
        context_schema = Schema([ColumnSchema("c1")])
        sequence_schema = Schema([ColumnSchema("s1"), ColumnSchema("s2")])
        with pytest.raises(ValueError) as exc_info:
            layer = BroadcastToSequence(context_schema, sequence_schema)
            inputs = {
                "c1": tf.constant([[1], [2]]),
                "s1": tf.constant([[[1, 2], [3, 4], [5, 6]], [[6, 3], [2, 3], [7, 3]]]),
                "s2": tf.constant([[[1, 2], [3, 4]], [[6, 3], [2, 3]]]),
            }
            layer(inputs)
        assert "All sequential features must share the same shape in the first two dims" in str(
            exc_info.value
        )

    def test_mask_propagation(self):
        masking_layer = tf.keras.layers.Masking(mask_value=0)
        partially_masked_inputs = {
            "a": tf.constant([[1], [2]]),
            "b": masking_layer(tf.constant([[[1], [0]], [[0], [1]]])),
        }
        context_schema = Schema([ColumnSchema("a")])
        sequence_schema = Schema([ColumnSchema("b")])
        broadcast_layer = BroadcastToSequence(context_schema, sequence_schema)
        outputs = broadcast_layer(partially_masked_inputs)
        self.assertAllEqual(outputs["a"]._keras_mask, tf.constant([[True, False], [False, True]]))
        self.assertAllEqual(outputs["b"]._keras_mask, tf.constant([[True, False], [False, True]]))

    def test_mask_propagation_ragged(self):
        masking_layer = tf.keras.layers.Masking(mask_value=0)
        partially_masked_inputs = {
            "a": tf.constant([[1], [2]]),
            "b": masking_layer(tf.ragged.constant([[[1]], [[0], [1]]])),
        }
        context_schema = Schema([ColumnSchema("a")])
        sequence_schema = Schema([ColumnSchema("b")])

        broadcast_layer = BroadcastToSequence(context_schema, sequence_schema)
        outputs = broadcast_layer(partially_masked_inputs)

        self.assertAllEqual(outputs["a"]._keras_mask, tf.ragged.constant([[True], [False, True]]))
        self.assertAllEqual(outputs["b"]._keras_mask, tf.ragged.constant([[True], [False, True]]))

    def test_in_model(self):
        masking_layer = tf.keras.layers.Masking(mask_value=0)
        partially_masked_inputs = {
            "a": tf.constant([[1], [2]]),
            "b": masking_layer(tf.ragged.constant([[[1]], [[0], [1]]])),
        }
        context_schema = Schema([ColumnSchema("a")])
        sequence_schema = Schema([ColumnSchema("b")])

        broadcast_layer = BroadcastToSequence(context_schema, sequence_schema)
        model = mm.Model(broadcast_layer)
        outputs = model(partially_masked_inputs)

        self.assertAllEqual(outputs["a"]._keras_mask, tf.ragged.constant([[True], [False, True]]))
        self.assertAllEqual(outputs["b"]._keras_mask, tf.ragged.constant([[True], [False, True]]))

    def test_compute_output_shape(self):
        masking_layer = tf.keras.layers.Masking(mask_value=0)
        partially_masked_inputs = {
            "a": tf.constant([[1], [2]]),
            "b": masking_layer(tf.ragged.constant([[[1]], [[0], [1]]])),
        }
        context_schema = Schema([ColumnSchema("a")])
        sequence_schema = Schema([ColumnSchema("b")])

        broadcast_layer = BroadcastToSequence(context_schema, sequence_schema)
        input_shape = {k: v.shape for k, v in partially_masked_inputs.items()}
        output_shape = broadcast_layer.compute_output_shape(input_shape)

        self.assertAllEqual(output_shape["a"], tf.TensorShape([2, None, None]))
        self.assertAllEqual(output_shape["b"], tf.TensorShape([2, None, None]))


@pytest.mark.parametrize(
    "only_selected_in_schema",
    [False, True],
)
def test_to_dense(only_selected_in_schema):
    test_case = TestCase()

    if only_selected_in_schema:
        schema = Schema(
            [
                create_categorical_column("feature1", tags=[Tags.CATEGORICAL], num_items=5),
                create_categorical_column("feature2", tags=[Tags.CATEGORICAL], num_items=5),
            ]
        )
        to_dense = mm.ToDense(schema.select_by_name("feature1"))
    else:
        # Converts all features to dense
        to_dense = mm.ToDense()

    feature1_dense = np.array([[1, 2, 3, 0], [0, 3, 1, 0]])
    feature2_dense = np.array([[2, 3, 4, 0], [2, 0, 1, 0]])

    data = {
        "feature1": tf.sparse.from_dense(feature1_dense),
        "feature2": tf.sparse.from_dense(feature2_dense),
    }

    output = to_dense(data)

    assert output["feature1"].shape == data["feature1"].shape
    assert output["feature2"].shape == data["feature2"].shape

    assert isinstance(output["feature1"], tf.Tensor)
    if only_selected_in_schema:
        assert isinstance(output["feature2"], tf.SparseTensor)
    else:
        assert isinstance(output["feature2"], tf.Tensor)

    # tf.convert_to_tensor(
    test_case.assertAllClose(output["feature1"], feature1_dense)
    if only_selected_in_schema:
        test_case.assertAllClose(tf.sparse.to_dense(output["feature2"]), feature2_dense)
    else:
        test_case.assertAllClose(output["feature2"], feature2_dense)


@pytest.mark.parametrize(
    "only_selected_in_schema",
    [False, True],
)
def test_to_sparse(only_selected_in_schema):
    if only_selected_in_schema:
        schema = Schema(
            [
                create_categorical_column("feature1", tags=[Tags.CATEGORICAL], num_items=5),
                create_categorical_column("feature2", tags=[Tags.CATEGORICAL], num_items=5),
            ]
        )
        to_sparse = mm.ToSparse(schema.select_by_name("feature1"))
    else:
        # Converts all features to dense
        to_sparse = mm.ToSparse()

    feature1_dense = np.array([[1, 2, 3, 0], [0, 3, 1, 0]])
    feature2_dense = np.array([[2, 3, 4, 0], [2, 0, 1, 0]])

    data = {
        "feature1": feature1_dense,
        "feature2": feature2_dense,
    }

    output = to_sparse(data)

    assert output["feature1"].shape == data["feature1"].shape
    assert output["feature2"].shape == data["feature2"].shape

    assert isinstance(output["feature1"], tf.SparseTensor)
    if only_selected_in_schema:
        assert isinstance(output["feature2"], tf.Tensor)
    else:
        assert isinstance(output["feature2"], tf.SparseTensor)

    # tf.convert_to_tensor(
    tf.debugging.assert_equal(tf.sparse.to_dense(output["feature1"]), feature1_dense)
    if only_selected_in_schema:
        tf.debugging.assert_equal(output["feature2"], feature2_dense)
    else:
        tf.debugging.assert_equal(tf.sparse.to_dense(output["feature2"]), feature2_dense)


def test_to_target_loader():
    schema = Schema(
        [
            ColumnSchema("a", tags=[Tags.USER, Tags.CATEGORICAL]),
            ColumnSchema("b", tags=[Tags.USER, Tags.CATEGORICAL]),
            ColumnSchema("c", tags=[Tags.ITEM, Tags.CATEGORICAL]),
        ]
    )

    input_df = pd.DataFrame(
        [
            {"a": 1, "b": 2, "c": 3},
            {"a": 4, "b": 5, "c": 6},
        ]
    )
    input_df = input_df[sorted(input_df.columns)]
    dataset = Dataset(input_df, schema=schema)

    # target is passed as a string.
    loader0 = mm.Loader(dataset, batch_size=10, transform=mm.ToTarget(schema, "c"))
    inputs0, targets0 = next(iter(loader0))
    assert sorted(inputs0.keys()) == ["a", "b"]
    assert targets0.numpy().tolist() == [[3], [6]]
    assert loader0.output_schema.select_by_tag(Tags.TARGET).column_names == ["c"]

    # target is passed as a ColumnSchema
    target_column_schema = ColumnSchema("c", tags=[Tags.CATEGORICAL])
    assert Tags.TARGET not in target_column_schema.tags
    loader1 = mm.Loader(dataset, batch_size=10, transform=mm.ToTarget(schema, target_column_schema))
    inputs1, targets1 = next(iter(loader1))
    assert sorted(inputs1.keys()) == ["a", "b"]
    assert targets1.numpy().tolist() == [[3], [6]]
    assert loader1.output_schema.select_by_tag(Tags.TARGET).column_names == ["c"]

    # target is passed as a Schema
    target_schema = schema.select_by_name("c")
    assert not target_schema.select_by_tag(Tags.TARGET)
    loader2 = mm.Loader(dataset, batch_size=10, transform=mm.ToTarget(schema, target_schema))
    inputs2, targets2 = next(iter(loader2))
    assert sorted(inputs2.keys()) == ["a", "b"]
    assert targets2.numpy().tolist() == [[3], [6]]
    assert loader2.output_schema.select_by_tag(Tags.TARGET).column_names == ["c"]

    # target is passed as a Tag
    loader3 = mm.Loader(dataset, batch_size=10, transform=mm.ToTarget(schema, Tags.ITEM))
    inputs3, targets3 = next(iter(loader3))
    assert sorted(inputs3.keys()) == ["a", "b"]
    assert targets3.numpy().tolist() == [[3], [6]]
    assert loader3.output_schema.select_by_tag(Tags.TARGET).column_names == ["c"]


def test_to_target_compute_output_schema():
    schema = Schema(
        [
            ColumnSchema("a", tags=[Tags.ITEM, Tags.CATEGORICAL]),
            ColumnSchema("b", tags=[Tags.USER, Tags.CATEGORICAL]),
            ColumnSchema("label", tags=[Tags.CATEGORICAL]),
        ]
    )
    to_target = mm.ToTarget(schema, "label")
    output_schema = to_target.compute_output_schema(schema)
    assert "label" in output_schema.select_by_tag(Tags.TARGET).column_names



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import pickle
import re
import shutil
from dataclasses import dataclass
from glob import glob
from pathlib import Path
from typing import List, Optional, Tuple, Union

import numpy as np
from tqdm import tqdm

import merlin.io
from merlin.core.dispatch import get_lib
from merlin.models.utils.example_utils import workflow_fit_transform
from merlin.models.utils.nvt_utils import require_nvt
from merlin.schema import Tags

try:
    import nvtabular as nvt

    Workflow = nvt.Workflow
except ImportError:
    Workflow = None


def get_aliccp(
    path: Union[str, Path],
    overwrite: bool = False,
    transformed_name: str = "transformed",
    nvt_workflow: Optional[Workflow] = None,
    file_size: int = 100000,
    **kwargs,
) -> Tuple[merlin.io.Dataset, merlin.io.Dataset]:
    """
    Load the Ali-CCP: Alibaba Click and Conversion Prediction dataset.

    We can't download this dataset automatically so this needs to be downloaded manually
    To use this function, prepare the data by following these 3 steps:
    1. Download the raw data from
    [tianchi.aliyun.com](https://tianchi.aliyun.com/dataset/dataDetail?dataId=408#1).
    2. Unzip the raw data to a directory.
    3. Run `prepare_alliccp(data_dir)` to convert the raw data to parquet files.

    Downloading & preparing the data can take quite a while.
    In case you want to use this dataset to run our tutorials, you can also opt for synthetic data.
    Synthetic data can be generated by running::
        from merlin.models.data.synthetic import generate_data

        data = generate_data("aliccp")

    Note, running this function requires the
    [nvtabular](https://github.com/NVIDIA-Merlin/NVTabular) library.

    Parameters
    ----------
    path: Union[str, Path]
        Directory to load the raw data from.
    overwrite: bool
        Whether to overwrite the transformed data.
    transformed_name: str
        Name of the transformed data.
        This is useful when you want to load the transformed data
        multiple times using different workflows.
    nvt_workflow: Optional[Workflow]
        Workflow to transform the raw data.
        If None, the default workflow will be used from `default_aliccp_transformation`.
    file_size: int
        Size of the parquet files to be loaded when preparing the data.

    Returns
    -------
    train: merlin.io.Dataset
        Training dataset.
    valid: merlin.io.Dataset
        Test dataset.
    """

    require_nvt()

    p = Path(path)
    raw_path = p / "raw"
    if not raw_path.exists():
        raw_path.mkdir(parents=True)
        prepare_alliccp(path, output_dir=raw_path, file_size=file_size, **kwargs)

    nvt_path = p / transformed_name
    train_path, valid_path = nvt_path / "train", nvt_path / "valid"
    nvt_path_exists = train_path.exists() and valid_path.exists()
    if not nvt_path_exists or overwrite:
        transform_aliccp(raw_path, nvt_path, nvt_workflow=nvt_workflow)

    train = merlin.io.Dataset(str(train_path), engine="parquet")
    valid = merlin.io.Dataset(str(valid_path), engine="parquet")

    return train, valid


def prepare_alliccp(
    data_dir: Union[str, Path],
    convert_train: bool = True,
    convert_test: bool = True,
    file_size: int = 100000,
    max_num_rows: Optional[int] = None,
    pickle_common_features=True,
    output_dir: Optional[Union[str, Path]] = None,
):
    """
    Convert Ali-CPP data to parquet files.

    To download the raw the Ali-CCP training and test datasets visit
    [tianchi.aliyun.com](https://tianchi.aliyun.com/dataset/dataDetail?dataId=408#1).

    Parameters
    ----------
    data_dir: Union[str, Path]
        Directory to load the raw data from.
    convert_train: bool
        Whether to convert the training data.
    convert_test: bool
        Whether to convert the test data.
    file_size: int
        Number of rows to write to each parquet file.
    max_num_rows: int, optional
        Maximum number of rows to read from the raw data.
    pickle_common_features: bool
        Whether to pickle the common features.
        When enabled it will make the conversion faster if it would be run again.
    output_dir: Union[str, Path], optional
        Directory to write the converted data to.
        If not specified the data will be written to the same directory as the raw data.

    Returns
    -------
    data_dir
    """

    if convert_train:
        _convert_data(
            str(data_dir),
            file_size,
            is_train=True,
            max_num_rows=max_num_rows,
            pickle_common_features=pickle_common_features,
            output_dir=output_dir,
        )
    if convert_test:
        _convert_data(
            str(data_dir),
            file_size,
            is_train=False,
            max_num_rows=max_num_rows,
            pickle_common_features=pickle_common_features,
            output_dir=output_dir,
        )

    return data_dir


def default_aliccp_transformation(add_target_encoding=True, **kwargs):
    import nvtabular as nvt
    from nvtabular import ops as nvt_ops

    cat = lambda: nvt_ops.Categorify(dtype="int32")  # noqa: E731

    user_id = ["user_id"] >> cat() >> nvt_ops.TagAsUserID()
    item_id = ["item_id"] >> cat() >> nvt_ops.TagAsItemID()

    item_features = (
        ["item_category", "item_shop", "item_brand", "item_intention"]
        >> cat()
        >> nvt_ops.TagAsItemFeatures()
    )

    user_features = (
        [
            "user_shops",
            "user_profile",
            "user_group",
            "user_gender",
            "user_age",
            "user_consumption_1",
            "user_consumption_2",
            "user_is_occupied",
            "user_geography",
            "user_intentions",
            "user_brands",
            "user_categories",
        ]
        >> cat()
        >> nvt_ops.TagAsUserFeatures()
    )

    user_item_features = (
        [
            "user_item_categories",
            "user_item_shops",
            "user_item_brands",
            "user_item_intentions",
        ]
        >> cat()
        >> nvt_ops.AddMetadata(tags=["user_item"])
    )

    context_features = ["position"] >> cat() >> nvt_ops.AddMetadata(tags=[Tags.CONTEXT])

    targets = ["click", "conversion"] >> nvt_ops.AddMetadata(
        tags=[Tags.BINARY_CLASSIFICATION, "target"]
    )
    outputs = (
        user_id
        + item_id
        + item_features
        + user_features
        + user_item_features
        + context_features
        + targets
    )

    if add_target_encoding:
        continuous = (
            ["user_id", "item_id"]
            >> nvt_ops.TargetEncoding(["click"], kfold=1, p_smooth=20)
            >> nvt_ops.Normalize()
        )
        outputs += continuous

    return nvt.Workflow(outputs)


def transform_aliccp(
    data: Union[str, Path, Tuple[merlin.io.Dataset, merlin.io.Dataset]],
    output_path: Union[str, Path],
    nvt_workflow=None,
    **kwargs,
):
    nvt_workflow = nvt_workflow or default_aliccp_transformation(**locals())

    if isinstance(data, (str, Path)):
        _train = glob(str(data / "train/*.parquet"))
        _valid = glob(str(data / "test/*.parquet"))
    elif (
        isinstance(data, tuple)
        and len(data) == 2
        and all(isinstance(x, merlin.io.Dataset) for x in data)
    ):
        _train, _valid = data
    else:
        raise ValueError("data must be a path or a tuple of train and valid datasets")

    workflow_fit_transform(nvt_workflow, _train, _valid, str(output_path), **kwargs)


@dataclass
class _Feature:
    name: str
    id: str
    tags: List[Union[str, Tags]]
    description: str


class _Features:
    def __init__(self):
        self.features: List[_Feature] = [
            # User
            _Feature("user_id", "101", [Tags.USER_ID, Tags.USER], "User ID"),
            _Feature(
                "user_categories",
                "109_14",
                [Tags.USER],
                "User historical behaviors of category ID and count",
            ),
            _Feature(
                "user_shops",
                "110_14",
                [Tags.USER],
                "User historical behaviors of shop ID and count",
            ),
            _Feature(
                "user_brands",
                "127_14",
                [Tags.USER],
                "User historical behaviors of brand ID and count",
            ),
            _Feature(
                "user_intentions",
                "150_14",
                [Tags.USER],
                "User historical behaviors of intention node ID and count",
            ),
            _Feature("user_profile", "121", [Tags.USER], "Categorical ID of User Profile"),
            _Feature("user_group", "122", [Tags.USER], "Categorical group ID of User Profile"),
            _Feature("user_gender", "124", [Tags.USER], "Users Gender ID"),
            _Feature("user_age", "125", [Tags.USER], "Users Age ID"),
            _Feature("user_consumption_1", "126", [Tags.USER], "Users Consumption Level Type I"),
            _Feature("user_consumption_2", "127", [Tags.USER], "Users Consumption Level Type II"),
            _Feature(
                "user_is_occupied", "128", [Tags.USER], "Users Occupation: whether or not to work"
            ),
            _Feature("user_geography", "129", [Tags.USER], "Users Geography Information"),
            # Item
            _Feature("item_id", "205", [Tags.ITEM, Tags.ITEM_ID], "Item ID"),
            _Feature(
                "item_category", "206", [Tags.ITEM], "Category ID to which the item belongs to"
            ),
            _Feature("item_shop", "207", [Tags.ITEM], "Shop ID to which item belongs to"),
            _Feature(
                "item_intention", "210", [Tags.ITEM], "Intention node ID which the item belongs to"
            ),
            _Feature("item_brand", "216", [Tags.ITEM], "Brand ID of the item"),
            # User-Item
            _Feature(
                "user_item_categories",
                "508",
                ["user_item"],
                "The combination of features with 109_14 and 206",
            ),
            _Feature(
                "user_item_shops",
                "509",
                ["user_item"],
                "The combination of features with 110_14 and 207",
            ),
            _Feature(
                "user_item_brands",
                "702",
                ["user_item"],
                "The combination of features with 127_14 and 216",
            ),
            _Feature(
                "user_item_intentions",
                "853",
                ["user_item"],
                "The combination of features with 150_14 and 210",
            ),
            # Context
            _Feature("position", "301", [Tags.CONTEXT], "A categorical expression of position"),
        ]

    @property
    def by_id(self):
        return {feature.id: feature.name for feature in self.features}


def _convert_common_features(common_path, pickle_path=None):
    common = {}

    with open(common_path, "r") as common_features:
        for csv_line in tqdm(common_features, desc="Reading common features..."):
            line = csv_line.strip().split(",")
            kv = np.array(re.split("[]", line[2]))
            keys = kv[range(0, len(kv), 3)]
            values = kv[range(1, len(kv), 3)]
            common[line[0]] = dict(zip(keys, values))

        if pickle_path:
            with open(pickle_path, "wb") as handle:
                pickle.dump(common, handle, protocol=pickle.HIGHEST_PROTOCOL)

    return common


# TODO: Optimize this function, right now it's too slow.
def _convert_data(
    data_dir,
    file_size,
    is_train=True,
    pickle_common_features=True,
    max_num_rows=None,
    output_dir=None,
):
    data_type = "train" if is_train else "test"
    output_dir = output_dir or data_dir

    common_path = os.path.join(data_dir, data_type, f"common_features_{data_type}.csv")
    path = os.path.join(data_dir, data_type, f"sample_skeleton_{data_type}.csv")
    common_features_path = os.path.join(data_dir, data_type, "common_features.pickle")

    common = {}
    if pickle_common_features:
        if os.path.exists(common_features_path):
            with open(common_features_path, "rb") as f:
                common = pickle.load(f)

    if not common:
        pickle_path = common_features_path if pickle_common_features else None
        common = _convert_common_features(common_path, pickle_path)

    current = []
    by_id = _Features().by_id

    out_dir = os.path.join(str(output_dir), data_type)
    tmp_dir = os.path.join(out_dir, "tmp")
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)

    with open(path, "r") as skeleton:
        for i, csv_line in tqdm(enumerate(skeleton), desc="Processing data..."):
            if max_num_rows and i >= max_num_rows:
                break

            line = csv_line.strip().split(",")
            if line[1] == "0" and line[2] == "1":
                continue
            kv = np.array(re.split("[]", line[5]))
            key = kv[range(0, len(kv), 3)]
            value = kv[range(1, len(kv), 3)]
            feat_dict = dict(zip(key, value))
            feat_dict.update(common[line[3]])
            feat_dict["click"] = int(line[1])
            feat_dict["conversion"] = int(line[2])

            current.append(feat_dict)

            if i > 0 and i % file_size == 0:
                df = get_lib().DataFrame(current)
                cols = []
                for col in list(df.columns):
                    if col == "click" or col == "conversion":
                        cols.append(col)
                    else:
                        cols.append(by_id[col])

                df.columns = cols
                index = int((i / file_size) - 1)
                df.to_parquet(
                    os.path.join(tmp_dir, f"{data_type}_{index}.parquet"),
                    overwrite=True,
                )
                current = []

    tmp_files = glob(os.path.join(tmp_dir, f"{data_type}_*.parquet"))
    dtypes = {f.name: "int32" for f in _Features().features}
    merlin.io.Dataset(tmp_files, dtypes=dtypes).to_parquet(out_dir)
    shutil.rmtree(tmp_dir)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inspect
from typing import Optional, Union

import tensorflow as tf
import transformers
from transformers import (
    AlbertConfig,
    BertConfig,
    GPT2Config,
    PretrainedConfig,
    RobertaConfig,
    TFPreTrainedModel,
    XLNetConfig,
)

from merlin.models.tf.core import combinators
from merlin.models.tf.core.base import Block, block_registry
from merlin.models.tf.transformers.transforms import PrepareTransformerInputs
from merlin.models.tf.utils.tf_utils import (
    maybe_deserialize_keras_objects,
    maybe_serialize_keras_objects,
)

TransformerBody = Union[TFPreTrainedModel, PretrainedConfig, tf.keras.layers.Layer]


def get_tf_main_layer(hf_model):
    """
    Extract serializable custom keras layer `TF*MainLayer` from the HuggingFace model
    """
    main_layer = [v for _, v in hf_model.__dict__.items() if isinstance(v, tf.keras.layers.Layer)][
        0
    ]
    return main_layer


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class TransformerBlock(Block):
    """
    Class to support HF Transformers for session-based and sequential-based recommendation models.
    Parameters
    ----------
    transformer: TransformerBody
        One of the following HuggingFace classes:
        Pretrained Config or model, or the custom keras layer `TF*MainLayer`,
        related to a specific transformer architecture.
    transformer_pre: tf.keras.layers.Layer
        Prepare the dictionary of inputs expected by the transformer layer
        by default PrepareTransformerInputs()
    transformer_post: [Union[str, tf.keras.layers.Layer]]
        Layer to extract the desired tensors from the HuggingFace dataclass
        `TFBaseModelOutputWithPoolingAndCrossAttentions`
        by default `last_hidden_state`
    pre: Optional[Union[str, tf.keras.layers.Layer]]
        A block to use before the main transformer block, by default None
    post: Optional[Union[str, tf.keras.layers.Layer]]
        A block to use after the main transformer block, by default None
    """

    def __init__(
        self,
        transformer: TransformerBody,
        pre: Optional[Union[str, tf.keras.layers.Layer]] = None,
        post: Optional[Union[str, tf.keras.layers.Layer]] = None,
        transformer_pre=PrepareTransformerInputs(),
        transformer_post: Optional[Union[str, tf.keras.layers.Layer]] = "last_hidden_state",
        **kwargs,
    ):
        super().__init__(**kwargs)

        if isinstance(transformer, PretrainedConfig):
            model_cls = transformers.TF_MODEL_MAPPING[transformer.__class__]
            self.transformer = get_tf_main_layer(model_cls(transformer))
        elif isinstance(transformer, TFPreTrainedModel):
            self.transformer = get_tf_main_layer(transformer)
        else:
            self.transformer = transformer

        if "transformer" in inspect.signature(transformer_pre.__init__).parameters:
            transformer_pre = transformer_pre(transformer=self.transformer)
        self.transformer_pre = transformer_pre
        if isinstance(transformer_post, str):
            transformer_post = block_registry.parse(transformer_post)
        self.transformer_post = transformer_post

        self.post = post
        self.pre = pre

    def build(self, input_shape=None):
        """Builds the sequential block

        Parameters
        ----------
        input_shape : tf.TensorShape, optional
            The input shape, by default None
        """
        combinators.build_sequentially(
            self, [*list(self.to_call_pre), self.transformer, *list(self.to_call_post)], input_shape
        )

    def call(self, inputs: tf.Tensor, **kwargs):
        """
        Parameters
        ----------
        inputs: tf.Tensor
            The 3D tensor of the sequence of interactions embeddings.
        """
        pre = combinators.call_sequentially(list(self.to_call_pre), inputs, **kwargs)
        transformer = self.transformer(pre)
        out = combinators.call_sequentially(list(self.to_call_post), transformer, **kwargs)

        return out

    @property
    def to_call_pre(self):
        if self.pre:
            yield self.pre

        yield self.transformer_pre

    @property
    def to_call_post(self):
        yield self.transformer_post

        if self.post:
            yield self.post

    def get_config(self):
        config = super().get_config()
        config = maybe_serialize_keras_objects(
            self,
            config,
            ["transformer", "pre", "post", "transformer_pre", "transformer_post"],
        )
        return config

    @classmethod
    def from_config(cls, config, custom_objects=None):
        config = maybe_deserialize_keras_objects(
            config,
            ["transformer", "pre", "post", "transformer_pre", "transformer_post"],
        )

        output = TransformerBlock(**config)
        output.__class__ = cls

        return output


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class BertBlock(TransformerBlock):
    """
    Class to prepare the configuration of a `Bert` model
    """

    def __init__(
        self,
        d_model: int,
        n_head: int,
        n_layer: int,
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        pre=None,
        post=None,
        transformer_pre=PrepareTransformerInputs(),
        transformer_post: Optional[Union[str, tf.keras.layers.Layer]] = "last_hidden_state",
        **kwargs,
    ):
        config = self.create_config(
            d_model,
            n_head,
            n_layer,
            hidden_act=hidden_act,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            dropout=dropout,
            pad_token=pad_token,
            log_attention_weights=log_attention_weights,
            **kwargs,
        )

        super().__init__(
            transformer=config,
            pre=pre,
            post=post,
            transformer_pre=transformer_pre,
            transformer_post=transformer_post,
        )

    def create_config(
        self,
        d_model,
        n_head,
        n_layer,
        max_position_embeddings=512,
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        **kwargs,
    ) -> BertConfig:

        return BertConfig(
            hidden_size=d_model,
            num_attention_heads=n_head,
            num_hidden_layers=n_layer,
            max_position_embeddings=max_position_embeddings,
            hidden_act=hidden_act,
            intermediate_size=d_model * 4,
            hidden_dropout_prob=dropout,
            attention_probs_dropout_prob=dropout,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            output_attentions=log_attention_weights,
            vocab_size=1,
            **kwargs,
        )


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class AlbertBlock(TransformerBlock):
    """
    Class to prepare the configuration of an `Albert` model
    """

    def __init__(
        self,
        d_model: int,
        n_head: int,
        n_layer: int,
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        transformer_pre=PrepareTransformerInputs(),
        transformer_post: Optional[Union[str, tf.keras.layers.Layer]] = "last_hidden_state",
        pre=None,
        post=None,
        **kwargs,
    ):
        config = self.create_config(
            d_model,
            n_head,
            n_layer,
            hidden_act=hidden_act,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            dropout=dropout,
            pad_token=pad_token,
            log_attention_weights=log_attention_weights,
            **kwargs,
        )

        super().__init__(
            config,
            pre=pre,
            post=post,
            transformer_pre=transformer_pre,
            transformer_post=transformer_post,
        )

    @classmethod
    def create_config(
        cls,
        d_model,
        n_head,
        n_layer,
        max_position_embeddings=512,
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        **kwargs,
    ) -> AlbertConfig:

        return AlbertConfig(
            hidden_size=d_model,
            num_attention_heads=n_head,
            num_hidden_layers=n_layer,
            hidden_act=hidden_act,
            intermediate_size=d_model * 4,
            hidden_dropout_prob=dropout,
            attention_probs_dropout_prob=dropout,
            max_position_embeddings=max_position_embeddings,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            output_attentions=log_attention_weights,
            vocab_size=1,
            **kwargs,
        )


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class RobertaBlock(TransformerBlock):
    """
    Class to prepare the configuration of a `RoBerta` model
    """

    def __init__(
        self,
        d_model,
        n_head,
        n_layer,
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        pre=None,
        post=None,
        transformer_pre=PrepareTransformerInputs(),
        transformer_post: Optional[Union[str, tf.keras.layers.Layer]] = "last_hidden_state",
        **kwargs,
    ):
        config = self.create_config(
            d_model,
            n_head,
            n_layer,
            hidden_act=hidden_act,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            dropout=dropout,
            pad_token=pad_token,
            log_attention_weights=log_attention_weights,
            **kwargs,
        )

        super().__init__(
            config,
            pre=pre,
            post=post,
            transformer_pre=transformer_pre,
            transformer_post=transformer_post,
        )

    @classmethod
    def create_config(
        cls,
        d_model,
        n_head,
        n_layer,
        max_position_embeddings=512,
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        **kwargs,
    ) -> RobertaConfig:

        return RobertaConfig(
            hidden_size=d_model,
            num_hidden_layers=n_layer,
            num_attention_heads=n_head,
            max_position_embeddings=max_position_embeddings + 8,
            hidden_act=hidden_act,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            dropout=dropout,
            pad_token_id=pad_token,
            output_attentions=log_attention_weights,
            vocab_size=1,
            **kwargs,
        )


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class XLNetBlock(TransformerBlock):
    """
    Class to prepare the configuration of a `XLNet` model
    """

    def __init__(
        self,
        d_model,
        n_head,
        n_layer,
        attn_type="bi",
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        mem_len=1,
        pre=None,
        post=None,
        transformer_pre=PrepareTransformerInputs(),
        transformer_post: Optional[Union[str, tf.keras.layers.Layer]] = "last_hidden_state",
        **kwargs,
    ):
        config = self.create_config(
            d_model,
            n_head,
            n_layer,
            attn_type=attn_type,
            hidden_act=hidden_act,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            dropout=dropout,
            pad_token=pad_token,
            log_attention_weights=log_attention_weights,
            mem_len=mem_len,
            **kwargs,
        )

        super().__init__(
            config,
            pre=pre,
            post=post,
            transformer_pre=transformer_pre,
            transformer_post=transformer_post,
        )

    @classmethod
    def create_config(
        cls,
        d_model,
        n_head,
        n_layer,
        attn_type="bi",
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        mem_len=1,
        **kwargs,
    ) -> XLNetConfig:

        return XLNetConfig(
            d_model=d_model,
            d_inner=d_model * 4,
            n_layer=n_layer,
            n_head=n_head,
            attn_type=attn_type,
            ff_activation=hidden_act,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            dropout=dropout,
            pad_token_id=pad_token,
            output_attentions=log_attention_weights,
            vocab_size=1,
            mem_len=mem_len,
            **kwargs,
        )


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class GPT2Block(TransformerBlock):
    """
    Class to prepare the configuration of a `GPT2` model
    """

    def __init__(
        self,
        d_model,
        n_head,
        n_layer,
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        pre=None,
        post=None,
        transformer_pre=PrepareTransformerInputs(),
        transformer_post: Optional[Union[str, tf.keras.layers.Layer]] = "last_hidden_state",
        **kwargs,
    ):
        config = self.create_config(
            d_model,
            n_head,
            n_layer,
            hidden_act=hidden_act,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            dropout=dropout,
            pad_token=pad_token,
            log_attention_weights=log_attention_weights,
            **kwargs,
        )

        super().__init__(
            config,
            pre=pre,
            post=post,
            transformer_pre=transformer_pre,
            transformer_post=transformer_post,
        )

    @classmethod
    def create_config(
        cls,
        d_model,
        n_head,
        n_layer,
        max_position_embeddings=512,
        hidden_act="gelu",
        initializer_range=0.01,
        layer_norm_eps=0.03,
        dropout=0.3,
        pad_token=0,
        log_attention_weights=False,
        **kwargs,
    ) -> GPT2Config:

        return GPT2Config(
            n_embd=d_model,
            n_inner=d_model * 4,
            n_layer=n_layer,
            n_head=n_head,
            activation_function=hidden_act,
            initializer_range=initializer_range,
            layer_norm_eps=layer_norm_eps,
            resid_pdrop=dropout,
            embd_pdrop=dropout,
            attn_pdrop=dropout,
            n_positions=max_position_embeddings,
            n_ctx=max_position_embeddings,
            output_attentions=log_attention_weights,
            vocab_size=1,
            **kwargs,
        )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import tensorflow as tf

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.loader import Loader
from merlin.models.tf.utils.testing_utils import assert_output_shape
from merlin.schema import Tags


@pytest.mark.parametrize("use_loader", [False, True])
def test_seq_predict_next(sequence_testing_data: Dataset, use_loader: bool):
    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    predict_next = mm.SequencePredictNext(schema=seq_schema, target=target, pre=mm.ListToRagged())

    batch = mm.sample_batch(sequence_testing_data, batch_size=8, include_targets=False)
    if use_loader:
        dataset_transformed = Loader(
            sequence_testing_data, batch_size=8, shuffle=False, transform=predict_next
        )
        output = next(iter(dataset_transformed))
    else:
        output = predict_next(batch)
    output_x, output_y = output
    output_y = output_y[target]

    as_ragged = mm.ListToRagged()
    batch = as_ragged(batch)

    # Checks if sequential input features were truncated in the last position
    for k, v in batch.items():
        if k in seq_schema.column_names:
            tf.Assert(tf.reduce_all(output_x[k] == v[:, :-1]), [output_x[k], v[:, :-1]])
        else:
            tf.Assert(tf.reduce_all(output_x[k] == v), [output_x[k], v])

    # Checks if the target is the shifted input feature
    tf.Assert(
        tf.reduce_all(output_y == batch[target][:, 1:]),
        [output_y, batch[target][:, 1:]],
    )


@pytest.mark.parametrize("use_loader", [False, True])
def test_seq_predict_last(sequence_testing_data: Dataset, use_loader: bool):
    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    predict_last = mm.SequencePredictLast(schema=seq_schema, target=target)

    batch = mm.sample_batch(sequence_testing_data, batch_size=8, include_targets=False)
    if use_loader:
        dataset_transformed = Loader(
            sequence_testing_data, batch_size=8, shuffle=False, transform=predict_last
        )
        output = next(iter(dataset_transformed))
    else:
        output = predict_last(batch)
    output_x, output_y = output
    output_y = output_y[target]

    as_ragged = mm.ListToRagged()
    batch = as_ragged(batch)

    # Checks if sequential input features were truncated in the last position
    for k, v in batch.items():
        if k in seq_schema.column_names:
            tf.Assert(tf.reduce_all(output_x[k] == v[:, :-1]), [output_x[k], v[:, :-1]])
        else:
            tf.Assert(tf.reduce_all(output_x[k] == v), [output_x[k], v])

    expected_target = tf.squeeze(batch[target][:, -1:].to_tensor(), axis=1)
    # Checks if the target is the last item
    tf.Assert(
        tf.reduce_all(output_y == expected_target),
        [output_y, expected_target],
    )


@pytest.mark.parametrize("use_loader", [False, True])
def test_seq_predict_random(sequence_testing_data: Dataset, use_loader: bool):
    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    predict_random = mm.SequencePredictRandom(schema=seq_schema, target=target)

    batch = mm.sample_batch(sequence_testing_data, batch_size=8, include_targets=False)
    if use_loader:
        dataset_transformed = Loader(
            sequence_testing_data, batch_size=8, shuffle=False, transform=predict_random
        )
        output = next(iter(dataset_transformed))
    else:
        output = predict_random(batch)
    output_x, output_y = output
    output_y = output_y[target]

    as_ragged = mm.ListToRagged()
    batch = as_ragged(batch)
    batch_size = batch[target].shape[0]

    for k, v in batch.items():

        if k in seq_schema.column_names:
            # Check if output sequences length is smaller than input sequences length
            tf.Assert(
                tf.reduce_all(output_x[k].row_lengths(1) < v.row_lengths(1)), [output_x[k], v]
            )
            # Check if first position of input and output sequence matches
            tf.Assert(
                tf.reduce_all(output_x[k][:, :1].to_tensor() == v[:, :1].to_tensor()),
                [output_x[k], v],
            )
        else:
            tf.Assert(tf.reduce_all(output_x[k] == v), [output_x[k], v])

    # Checks if the target has the right shape
    tf.Assert(tf.reduce_all(tf.shape(output_y) == batch_size), [])


def test_seq_predict_next_output_shape(sequence_testing_data):
    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    predict_next = mm.SequencePredictNext(schema=seq_schema, target=target)

    batch = mm.sample_batch(sequence_testing_data, batch_size=8, include_targets=False)

    input_shapes = dict()
    expected_output_shapes = dict()
    for k, v in batch.items():
        if k in seq_schema.column_names:
            input_shapes[k] = (v[0].shape, v[1].shape)
            expected_output_shapes[k] = tf.TensorShape([v[1].shape[0], None])
        else:
            input_shapes[k] = expected_output_shapes[k] = v.shape

    # Hacking one input feature to have defined seq length
    input_shapes["categories"] = tf.TensorShape([8, 4])
    expected_output_shapes["categories"] = tf.TensorShape([8, 3])

    output_shape = predict_next.compute_output_shape(input_shapes)
    assert_output_shape(output_shape, expected_output_shapes)


def test_seq_predict_next_serialize_deserialize(sequence_testing_data):
    predict_next = mm.SequencePredictNext(sequence_testing_data.schema, "item_id_seq")
    assert isinstance(predict_next.from_config(predict_next.get_config()), mm.SequencePredictNext)


def asserts_mlm_target_mask(target_mask):
    # Checks if there is no sequence with no elements masked as target
    tf.assert_equal(
        tf.reduce_all(tf.reduce_any(target_mask, axis=1)),
        True,
        message=f"There are sequences with no targets masked {target_mask.numpy()}",
    )
    # Checks if there is no sequence with all elements masked
    tf.assert_equal(
        tf.reduce_any(tf.reduce_all(target_mask, axis=1)),
        False,
        message=f"There are sequences with all targets masked {target_mask.numpy()}",
    )


def test_seq_random_masking(sequence_testing_data: Dataset):
    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    predict_masked = mm.SequenceMaskRandom(schema=seq_schema, target=target, masking_prob=0.3)

    batch = mm.sample_batch(sequence_testing_data, batch_size=8, include_targets=False)

    output = predict_masked(batch)
    output_x, output_y = output.outputs, output.targets
    output_y = output_y[target]

    tf.Assert(tf.reduce_all(output_y == output_x[target]), [output_y, output_x[target]])

    target_mask = output_y._keras_mask

    asserts_mlm_target_mask(target_mask)

    as_ragged = mm.ListToRagged()
    batch = as_ragged(batch)

    for k, v in batch.items():
        # Checking if inputs values didn't change
        tf.Assert(tf.reduce_all(output_x[k] == v), [output_x[k], v])

        # Checks if for sequential input columns the mask has been assigned
        # (opposite of the target mask)
        if k in seq_schema.column_names:
            tf.Assert(
                tf.reduce_all(output_x[k]._keras_mask == target_mask),
                [],
            )


def test_seq_predict_masked_target_not_present(sequence_testing_data: Dataset):
    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE)
    with pytest.raises(ValueError) as exc_info:
        _ = mm.SequenceMaskRandom(schema=seq_schema, target="NOT_EXISTS", masking_prob=0.3)
    assert "The target column needs to be part of the sequential schema" in str(exc_info.value)


def test_seq_predict_masked_serialize_deserialize(sequence_testing_data):
    predict_masked = mm.SequenceMaskRandom(sequence_testing_data.schema, "item_id_seq")
    assert isinstance(
        predict_masked.from_config(predict_masked.get_config()), mm.SequenceMaskRandom
    )


@pytest.mark.parametrize("dense", [False, True])
@pytest.mark.parametrize("target_as_dict", [False, True])
def test_seq_mask_random_replace_embeddings(
    sequence_testing_data: Dataset, dense: bool, target_as_dict: bool
):
    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE).select_by_name(
        ["item_id_seq", "categories"]
    )

    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    predict_masked = mm.SequenceMaskRandom(schema=seq_schema, target=target, masking_prob=0.3)

    batch = mm.sample_batch(sequence_testing_data, batch_size=8, include_targets=False)

    output = predict_masked(batch)
    inputs, targets = output.outputs, output.targets
    targets = targets[target]

    emb = tf.keras.layers.Embedding(1000, 16)
    item_id_emb_seq = emb(inputs["item_id_seq"])
    targets_mask = targets._keras_mask
    if dense:
        item_id_emb_seq = item_id_emb_seq.to_tensor()
        targets = targets.to_tensor()
        targets_mask = targets_mask.to_tensor()
        targets._keras_mask = targets_mask

    if target_as_dict:
        # Making targets different in dict, the valid one is "target2" which is 2D
        targets = {"target1": tf.ragged.constant([1, 2, 3, 4, 5, 6, 7, 8]), "target2": targets}

    masked_embeddings = mm.ReplaceMaskedEmbeddings()
    output = masked_embeddings(item_id_emb_seq, targets=targets, training=True)

    replaced_mask = tf.logical_not(tf.reduce_all(output == item_id_emb_seq, axis=2))

    if not dense:
        replaced_mask = replaced_mask.with_row_splits_dtype(tf.int64)
    tf.Assert(tf.reduce_all(replaced_mask == targets_mask), [])
    asserts_mlm_target_mask(replaced_mask)


def test_replace_masked_input_embeddings_no_target():
    item_id_emb_seq = tf.random.uniform((8, 10), dtype=tf.float32)
    item_id_emb_seq._keras_mask = tf.cast(
        tf.random.uniform((8,), minval=0, maxval=2, dtype=tf.int32), tf.bool
    )
    targets = None

    masked_embeddings = mm.ReplaceMaskedEmbeddings()
    output = masked_embeddings(item_id_emb_seq, targets=targets)
    # Checks that no input embedding was replaced, as there was no masking defined
    tf.Assert(tf.logical_not(tf.reduce_all(output == item_id_emb_seq)), [])


def test_not_replace_unmasked_sequence_embeddings():
    item_id_emb_seq = tf.random.uniform((8, 10), dtype=tf.float32)
    targets = tf.random.uniform((8, 10), dtype=tf.float32)

    masked_embeddings = mm.ReplaceMaskedEmbeddings()
    output = masked_embeddings(item_id_emb_seq, targets=targets)
    # Checks that no input embedding was replaced, as there was no masking defined
    tf.Assert(tf.reduce_all(output == item_id_emb_seq), [])


def test_replace_masked_input_2d_embeddings_incompatible_2d_mask():
    item_id_emb_seq = tf.random.uniform((8, 10), dtype=tf.float32)
    item_id_emb_seq._keras_mask = tf.cast(
        tf.random.uniform((8, 10), minval=0, maxval=2, dtype=tf.int32), tf.bool
    )
    masked_embeddings = mm.ReplaceMaskedEmbeddings()

    with pytest.raises(Exception) as exc_info:
        _ = masked_embeddings(item_id_emb_seq)
    assert "The inputs and mask need to be compatible" in str(exc_info.value)


def test_replace_masked_input_2d_embeddings_incompatible_ragged_2d_mask():
    item_id_emb_seq = tf.random.uniform((8, 10), dtype=tf.float32)
    item_id_emb_seq._keras_mask = tf.RaggedTensor.from_tensor(
        tf.cast(tf.random.uniform((8, 10), minval=0, maxval=2, dtype=tf.int32), tf.bool)
    )
    masked_embeddings = mm.ReplaceMaskedEmbeddings()

    with pytest.raises(Exception) as exc_info:
        _ = masked_embeddings(item_id_emb_seq)
    assert "The inputs and mask need to be compatible" in str(exc_info.value)


def test_seq_replace_embeddings_ragged_tensor_last_dim_none():
    inputs = tf.ragged.constant([[[1, 2]], [[3]]])

    masked_embeddings = mm.ReplaceMaskedEmbeddings()
    with pytest.raises(Exception) as exc_info:
        _ = masked_embeddings(inputs, training=True)
    assert "The last dim of inputs cannot be None" in str(exc_info.value)
